#!python3
"""
Mosaic GUI implementation

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""
import os
from typing import List
from os.path import extsep, basename

import vtk
import numpy as np
from qtpy.QtCore import Qt, QEvent, QSize, QTimer, QThread
from qtpy.QtWidgets import (
    QApplication,
    QMainWindow,
    QVBoxLayout,
    QWidget,
    QSplitter,
    QFileDialog,
    QMenu,
    QHBoxLayout,
    QPushButton,
    QDockWidget,
    QButtonGroup,
    QShortcut,
    QMessageBox,
    QCheckBox,
)
from qtpy.QtGui import (
    QAction,
    QGuiApplication,
    QActionGroup,
    QKeyEvent,
    QDropEvent,
    QCursor,
    QDragEnterEvent,
)
import qtawesome as qta
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor

from .data import MosaicData
from .settings import Settings
from .animate import ExportManager
from .parallel import BackgroundTaskManager
from .tabs import SegmentationTab, ModelTab, IntelligenceTab
from .dialogs import (
    TiltControlDialog,
    KeybindsDialog,
    ImportDataDialog,
    ProgressDialog,
    AppSettingsDialog,
)
from .widgets import (
    MultiVolumeViewer,
    AxesWidget,
    RibbonToolBar,
    TrajectoryPlayer,
    LegendWidget,
    ScaleBarWidget,
    ObjectBrowserSidebar,
    ViewerModes,
    StatusIndicator,
    CursorModeHandler,
    BoundingBoxManager,
)


class ImportThread(QThread):
    def run(self):
        from .meshing.utils import to_open3d
        from .parametrization import TriangularMesh
        from .dialogs import PropertyAnalysisDialog
        from .parametrization import PARAMETRIZATION_TYPE


class App(QMainWindow):

    def __init__(self):
        super().__init__()
        self.setWindowState(Qt.WindowNoState)

        central_widget = QWidget()
        self.setCentralWidget(central_widget)
        layout = QVBoxLayout(central_widget)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        # Render Block
        self.vtk_widget = QVTKRenderWindowInteractor()

        self.cdata = MosaicData(self.vtk_widget)

        self.renderer = vtk.vtkRenderer()
        self.render_window = self.vtk_widget.GetRenderWindow()
        self.render_window.AddRenderer(self.renderer)
        self.apply_render_settings()

        # Setup GUI interactions
        self.interactor = self.vtk_widget.GetRenderWindow().GetInteractor()
        self.interactor.Initialize()
        self.interactor.AddObserver("RightButtonPressEvent", self.on_right_click)
        self.interactor.AddObserver("KeyPressEvent", self.on_key_press)
        self.interactor.SetDesiredUpdateRate(Settings.rendering.target_fps)

        self.tab_bar = QWidget()
        self.tab_bar.setFixedHeight(40)
        self.tab_bar.setStyleSheet(
            """
            QWidget {
                border-bottom: 1px solid #6b7280;
            }
        """
        )
        tab_layout = QHBoxLayout(self.tab_bar)
        tab_layout.setContentsMargins(16, 0, 16, 0)
        tab_layout.setSpacing(4)

        self.tab_button_group = QButtonGroup(self)
        self.tab_button_group.setExclusive(True)

        self.setup_widgets()
        self.tab_buttons = {}
        self.tab_ribbon = RibbonToolBar(self)
        data = {"cdata": self.cdata, "ribbon": self.tab_ribbon, "legend": self.legend}

        self.tabs = [
            (SegmentationTab(**data), "Segmentation"),
            (ModelTab(**data), "Parametrization"),
            (IntelligenceTab(**data), "Intelligence"),
            # (DevelopmentTab(**data), "Development"),
        ]

        for index, (tab, name) in enumerate(self.tabs):
            btn = QPushButton(name)
            btn.setObjectName("TabButton")
            btn.setProperty("tab_id", index)
            btn.setCheckable(True)
            self.tab_button_group.addButton(btn, index)

            btn.setStyleSheet(
                """
                QPushButton {
                    border: none;
                    padding: 11px 24px;
                    font-size: 13px;
                    border-bottom: 2px solid transparent;
                    min-width: 100px;
                }
                QPushButton:hover:!checked {
                    font-weight: 500;
                }
                QPushButton:checked {
                    font-weight: 500;
                    color: rgba(99, 102, 241, 1.0);
                    border-bottom: 2px solid rgba(99, 102, 241, 1.0);
                }
                QPushButton:focus {
                    outline: none;
                }
            """
            )
            tab_layout.addWidget(btn)
            self.tab_buttons[index] = btn

        self.tab_button_group.idClicked.connect(
            lambda tab_id: self.tabs[tab_id][0].show_ribbon()
        )

        tab_layout.addStretch()
        self.tab_buttons[0].setChecked(True)
        self.tabs[0][0].show_ribbon()

        layout.addWidget(self.tab_bar)
        layout.addWidget(self.tab_ribbon)

        list_wrapper = ObjectBrowserSidebar()
        list_wrapper.set_title("Object Browser")
        list_wrapper.add_widget("cluster", "Cluster", self.cdata.data.data_list)
        list_wrapper.add_widget("model", "Model", self.cdata.models.data_list)

        splitter = QSplitter(Qt.Orientation.Horizontal)
        splitter.addWidget(list_wrapper)
        splitter.addWidget(self.vtk_widget)
        splitter.setSizes([200, self.width() - 200])
        splitter.setStretchFactor(0, 0)
        splitter.setStretchFactor(1, 1)

        layout.addWidget(splitter)

        self.actor_collection = vtk.vtkActorCollection()
        self.setup_menu()

        self._import_thread = ImportThread()
        self._import_thread.start()

        self.escape_shortcut = QShortcut(Qt.Key.Key_Escape, self.vtk_widget)
        self.escape_shortcut.activated.connect(self.handle_escape_key)

        QTimer.singleShot(2000, self._check_for_updates)

        self.setAcceptDrops(True)
        self._drag_active = False

    def dragEnterEvent(self, event: QDragEnterEvent):
        if event.mimeData().hasUrls():
            valid_files = False
            for url in event.mimeData().urls():
                if url.isLocalFile():
                    valid_files = True
                    break

            if valid_files:
                event.acceptProposedAction()
                self._drag_active = True
                QApplication.setOverrideCursor(QCursor(Qt.DragCopyCursor))
                self.setStyleSheet(
                    self.styleSheet()
                    + """
                    QMainWindow {
                        border: 2px dashed rgba(99, 102, 241, 0.3);
                    }
                    """
                )
        return super().dragEnterEvent(event)

    def dragLeaveEvent(self, event):
        self._drag_active = False
        QApplication.restoreOverrideCursor()
        self._update_style()

    def dropEvent(self, event: QDropEvent):
        self._drag_active = False
        QApplication.restoreOverrideCursor()
        self._update_style()

        if not event.mimeData().hasUrls():
            event.ignore()
            return None

        event.acceptProposedAction()

        file_paths = []
        for url in event.mimeData().urls():
            if url.isLocalFile():
                file_paths.append(url.toLocalFile())

        if not file_paths:
            return None

        session_files = [f for f in file_paths if f.lower().endswith(".pickle")]
        data_files = [f for f in file_paths if not f.lower().endswith(".pickle")]

        if session_files:
            if len(session_files) > 1:
                QMessageBox.warning(
                    self,
                    "Multiple Session Files",
                    "Only one session file can be loaded at a time. ",
                )
            self._load_session(session_files[0])

        if len(data_files):
            self._open_files(data_files)

    def sizeHint(self):
        """Provide the preferred size for the main window."""
        screen = QGuiApplication.primaryScreen().geometry()
        width = int(screen.width() * 0.9)
        height = int(screen.height() * 0.9)
        return QSize(width, height)

    def show(self):
        """Override show to position after Qt sizes the window."""
        self.resize(self.sizeHint())
        super().show()

        # Position after showing (when size is established)
        screen = QGuiApplication.primaryScreen().geometry()
        x = (screen.width() - self.width()) // 2
        y = (screen.height() - self.height()) // 2
        self.move(x, y)

    def apply_render_settings(self):
        self.renderer.SetBackground(
            *[float(x) for x in Settings.rendering.background_color]
        )
        self.renderer_next_background = [
            float(x) for x in Settings.rendering.background_color_alt
        ]

        # Check how these settings perform
        self.renderer.GradientBackgroundOff()
        self.renderer.SetUseDepthPeeling(Settings.rendering.use_depth_peeling)
        self.renderer.SetOcclusionRatio(Settings.rendering.occlusion_ratio)
        self.renderer.SetMaximumNumberOfPeels(Settings.rendering.max_depth_peels)
        self.renderer.SetUseFXAA(Settings.rendering.enable_fxaa)

        self.render_window.SetMultiSamples(Settings.rendering.multisamples)
        self.render_window.SetPointSmoothing(Settings.rendering.point_smoothing)
        self.render_window.SetLineSmoothing(Settings.rendering.line_smoothing)
        self.render_window.SetPolygonSmoothing(Settings.rendering.polygon_smoothing)
        self.render_window.SetDesiredUpdateRate(Settings.rendering.target_fps)
        self.render_window.Render()

        if not hasattr(self, "cdata"):
            return None

        from .actor import ActorFactory

        if not ActorFactory().is_synced():
            ActorFactory().update_from_settings()
            self.cdata.refresh_actors()

    def handle_escape_key(self, *args, **kwargs):
        """Handle escape key press - switch to viewing mode if not already in it."""
        self._transition_modes(self.cursor_handler.current_mode)
        self.interactor.SetInteractorStyle(vtk.vtkInteractorStyleTrackballCamera())

    def on_key_press(self, obj, event):
        key = obj.GetKeyCode()

        if key in ["x", "c", "z"]:
            self.set_camera_view(key)
        elif key == "v":
            self.swap_camera_view_direction(key)
        elif key in ["d"]:
            current_color = self.renderer.GetBackground()
            self.renderer.SetBackground(*self.renderer_next_background)
            self.renderer_next_background = current_color
            self.vtk_widget.GetRenderWindow().Render()
        elif key in ["\x7f", "\x08"]:
            self.cdata.data.remove()
        elif key == "m":
            self.cdata.data.merge()
        elif key == "e":
            self.cdata.highlight_clusters_from_selected_points()
        elif key == "h":
            self.cdata.visibility_unselected(visible=False)
        elif key == "H":
            self.cdata.visibility_unselected(visible=True)
        elif key == "s":
            self._transition_modes(ViewerModes.VIEWING)
            self.cdata.swap_area_picker()
            self.toggle_selection_menu()
        elif key == "E":
            self._transition_modes(ViewerModes.PICKING)
        elif key == "a":
            self._transition_modes(ViewerModes.DRAWING)
        elif key == "A":
            self._transition_modes(ViewerModes.CURVE)
        elif key == "q":
            self._transition_modes(ViewerModes.MESH_DELETE)
        elif key == "Q":
            self._transition_modes(ViewerModes.MESH_ADD)
        elif key == "r":
            self._transition_modes(ViewerModes.SELECTION)

    def on_right_click(self, obj, event):
        self.cdata.data.deselect()
        self.cdata.models.deselect()

    def _transition_modes(self, new_mode):
        current_mode = self.cursor_handler.current_mode
        if current_mode in (
            ViewerModes.MESH_ADD,
            ViewerModes.MESH_DELETE,
            ViewerModes.CURVE,
        ):
            current_style = self.interactor.GetInteractorStyle()
            if hasattr(current_style, "cleanup"):
                current_style.cleanup()

            self.cdata.swap_area_picker()
            self.cdata.swap_area_picker()

        self.cdata.activate_viewing_mode()
        self.status_indicator.update_status(interaction=new_mode.value)
        if current_mode == new_mode:
            self.status_indicator.update_status(interaction=ViewerModes.VIEWING.value)
            return self.cursor_handler.update_mode(ViewerModes.VIEWING)

        if new_mode == ViewerModes.DRAWING:
            self.cdata.data.activate_drawing_mode()
        elif new_mode == ViewerModes.CURVE:
            from .styles import CurveBuilderInteractorStyle

            style = CurveBuilderInteractorStyle(self, self.cdata)
            self.interactor.SetInteractorStyle(style)
            style.SetDefaultRenderer(self.renderer)
        elif new_mode == ViewerModes.SELECTION:
            self.interactor.SetInteractorStyle(vtk.vtkInteractorStyleRubberBandPick())
        elif new_mode == ViewerModes.PICKING:
            self.cdata.activate_picking_mode()
        elif new_mode in (ViewerModes.MESH_ADD, ViewerModes.MESH_DELETE):
            from .styles import MeshEditInteractorStyle

            style = MeshEditInteractorStyle(self, self.cdata)
            self.interactor.SetInteractorStyle(style)
            style.SetDefaultRenderer(self.renderer)
            if new_mode == ViewerModes.MESH_ADD:
                style.toggle_add_face_mode()

        return self.cursor_handler.update_mode(new_mode)

    def set_camera_view(
        self, view_key, aligned_direction=True, elevation=0, azimuth=0, pitch=0
    ):
        camera = self.renderer.GetActiveCamera()
        focal_point = camera.GetFocalPoint()
        position = camera.GetPosition()

        distance = np.linalg.norm(np.subtract(position, focal_point))
        distance = distance if aligned_direction else -distance
        if view_key == "z":
            view = (1, 0, 1)
            position_vec = (0, 0, 1)
        elif view_key == "c":
            view = (1, 0, 0)
            position_vec = (0, 1, 0)
        elif view_key == "x":
            view = (0, 1, 0)
            position_vec = (1, 0, 0)
        else:
            return -1

        transform = vtk.vtkTransform()
        transform.Identity()
        transform.RotateWXYZ(elevation, *(0, 0, 1))
        transform.RotateWXYZ(azimuth, *(0, 1, 0))
        transform.RotateWXYZ(pitch, *(1, 0, 0))

        view = transform.TransformVector(view)
        position_vec = np.array(transform.TransformVector(position_vec))
        position_vec /= np.linalg.norm(position_vec)
        position_vec *= distance

        position = np.add(focal_point, position_vec)
        current_view = getattr(self, "_camera_view", None)
        if current_view != view_key:
            focal_point = (0, 0, 0)
            position = position_vec

        camera.SetPosition(*position)
        camera.SetViewUp(*view)
        camera.SetFocalPoint(*focal_point)
        if current_view != view_key:
            self.renderer.ResetCamera()

        self._camera_view = view_key
        self._camera_elevation = elevation
        self._camera_azimuth = azimuth
        self._camera_pitch = pitch
        self._camera_direction = aligned_direction
        self.vtk_widget.GetRenderWindow().Render()

    def reset_camera_view(self):
        self._camera_view = None
        self.set_camera_view("z")

    def swap_camera_view_direction(self, view_key):
        view = getattr(self, "_camera_view", None)
        if view is None:
            return -1

        direction = getattr(self, "_camera_direction", True)
        return self.set_camera_view(view, not direction)

    def _update_style(self):
        self.setStyleSheet(
            """
            QMenuBar {
                border-bottom: 1px solid #6b7280;
            }
            QMenuBar::item {
                padding: 4px 8px;
            }
            QMenuBar::item:selected {
                background-color: #1a000000;
                border-radius: 4px;
            }
            QMenu {
                border-radius: 4px;
                padding: 4px;
            }
            QMenu::item {
                padding: 4px 24px 4px 8px;
                border-radius: 4px;
            }
            QMenu::item:selected {
                background-color: #1a000000;
            }
        """
        )

    def changeEvent(self, event):
        if event.type() == QEvent.Type.PaletteChange:
            self._update_style()
        super().changeEvent(event)

    def setup_widgets(self):
        self.legend = LegendWidget(self.renderer, self.interactor)

        self.volume_dock = None
        self.volume_viewer = MultiVolumeViewer(self.vtk_widget, legend=self.legend)
        self.cdata.data.render_update.connect(
            self.volume_viewer.primary.handle_projection_change
        )
        self.cdata.models.render_update.connect(
            self.volume_viewer.primary.handle_projection_change
        )

        self.cursor_handler = CursorModeHandler(self.vtk_widget)
        self.axes_widget = AxesWidget(self.renderer, self.interactor)
        self.trajectory_player = TrajectoryPlayer(self.cdata)
        self.scale_bar = ScaleBarWidget(self.renderer, self.interactor)
        self.export_manager = ExportManager(
            self.vtk_widget, self.volume_viewer, self.cdata
        )
        self.status_indicator = StatusIndicator(self)

        self.bbox_manager = BoundingBoxManager(
            self.renderer, self.interactor, self.cdata
        )

        task_manager = BackgroundTaskManager.instance()
        task_manager.running_tasks.connect(
            lambda n: self.status_indicator.update_status(busy=n >= 1)
        )
        task_manager.task_started.connect(
            lambda _, name: self.status_indicator.update_status(busy=True, task=name)
        )

        self._setup_volume_viewer()
        self._setup_trajectory_player()

    def setup_menu(self):
        self._update_style()

        menu_bar = self.menuBar()

        file_menu = menu_bar.addMenu("File")
        view_menu = menu_bar.addMenu("View")
        interact_menu = menu_bar.addMenu("Actions")
        preference_menu = menu_bar.addMenu("Preferences")
        help_menu = menu_bar.addMenu("Help")

        # File menu actions
        new_session_action = QAction("Load Session", self)
        new_session_action.triggered.connect(self.load_session)
        new_session_action.setShortcut("Ctrl+N")

        add_file_action = QAction("Open", self)
        add_file_action.triggered.connect(self.open_files)
        add_file_action.setShortcut("Ctrl+O")

        undo_action = QAction("Undo", self)
        undo_action.triggered.connect(self.cdata.data.undo)
        undo_action.setShortcut("Ctrl+Z")

        save_file_action = QAction("Save Session", self)
        save_file_action.triggered.connect(self.save_session)
        save_file_action.setShortcut("Ctrl+S")

        close_file_action = QAction("Close Session", self)
        close_file_action.triggered.connect(self.close_session)

        self.recent_file_actions = []
        self.recent_menu = QMenu("Recent Files", self)
        for i in range(Settings.ui.max_recent_files):
            action = QAction(self)
            action.setVisible(False)
            action.triggered.connect(self._open_recent_file)
            self.recent_file_actions.append(action)
            self.recent_menu.addAction(action)

        self.update_recent_files_menu()

        quit_action = QAction("Quit", self)
        quit_action.setShortcut("Ctrl+Q")
        quit_action.triggered.connect(self.close)

        screenshot_action = QAction("Save Viewer Screenshot", self)
        screenshot_action.triggered.connect(
            lambda x: self.export_manager.save_screenshot()
        )
        screenshot_action.setShortcut("Ctrl+P")

        animation_action = QAction("Export Animation", self)
        animation_action.triggered.connect(
            lambda x: self.export_manager.export_animation()
            # lambda x: self._animate()
        )
        animation_action.setShortcut("Ctrl+E")

        clipboard_action = QAction("Viewer Screenshot to Clipboard", self)
        clipboard_action.triggered.connect(
            lambda x: self.export_manager.copy_screenshot_to_clipboard()
        )
        clipboard_action.setShortcut("Ctrl+Shift+C")

        clipboard_window_action = QAction("Window Screenshot to Clipboard", self)
        clipboard_window_action.triggered.connect(
            lambda x: self.export_manager.copy_screenshot_to_clipboard(window=True)
        )
        clipboard_window_action.setShortcut("Ctrl+Shift+W")

        # Setup axes control menu
        axes_menu = QMenu("Axes", self)
        visible_action = QAction("Visible", self)
        visible_action.setCheckable(True)
        visible_action.setChecked(self.axes_widget.visible)
        visible_action.triggered.connect(
            lambda checked: (
                self.axes_widget.set_visibility(checked),
                self.vtk_widget.GetRenderWindow().Render(),
            )
        )
        labels_action = QAction("Labels", self)
        labels_action.setCheckable(True)
        labels_action.setChecked(self.axes_widget.labels_visible)
        labels_action.triggered.connect(
            lambda checked: (
                self.axes_widget.set_labels_visible(checked),
                self.vtk_widget.GetRenderWindow().Render(),
            )
        )
        colored_action = QAction("Colored", self)
        colored_action.setCheckable(True)
        colored_action.setChecked(self.axes_widget.colored)
        colored_action.triggered.connect(
            lambda checked: (
                self.axes_widget.set_colored(checked),
                self.vtk_widget.GetRenderWindow().Render(),
            )
        )
        arrow_action = QAction("Arrows", self)
        arrow_action.setCheckable(True)
        arrow_action.setChecked(self.axes_widget.arrow_heads_visible)
        arrow_action.triggered.connect(
            lambda checked: (
                self.axes_widget.set_arrow_heads_visible(checked),
                self.vtk_widget.GetRenderWindow().Render(),
            )
        )
        axes_menu.addAction(visible_action)
        axes_menu.addAction(labels_action)
        axes_menu.addAction(colored_action)
        axes_menu.addAction(arrow_action)

        # Handle different camera angles
        tilt_menu = QMenu("Camera", self)
        self.tilt_dialog = TiltControlDialog(self)
        show_tilt_control = QAction(
            qta.icon("fa5s.sliders-h", opacity=0.7, color="gray"),
            "Tilt Controls...",
            self,
        )
        show_tilt_control.triggered.connect(self.tilt_dialog.show)
        tilt_menu.addAction(show_tilt_control)

        tilt_menu.addSeparator()
        tilt_group = QActionGroup(self)
        tilt_group.setExclusive(True)
        for angle in [0, 15, 30, 45, 60, 90]:
            action = QAction(f"{angle}°", self)
            action.triggered.connect(
                lambda checked, a=angle: self.set_camera_view(
                    getattr(self, "_camera_view", "x"),
                    getattr(self, "_camera_direction", True),
                    view_angle=a,
                )
            )
            tilt_menu.addAction(action)

        tilt_menu.addSeparator()
        reset_action = QAction(
            qta.icon("fa5s.undo", opacity=0.7, color="gray"), "Reset Tilt", self
        )
        reset_action.setShortcut("Ctrl+T")
        reset_action.triggered.connect(self.tilt_dialog.reset_tilt)
        tilt_menu.addAction(reset_action)

        coloring_menu = QMenu("Coloring", self)
        coloring_group = QActionGroup(self)
        coloring_group.setExclusive(True)

        self.color_default_action = QAction("Default", self)
        self.color_default_action.setCheckable(True)
        self.color_default_action.setChecked(True)
        self.color_default_action.triggered.connect(
            lambda: self.cdata.set_coloring_mode("default")
        )
        coloring_group.addAction(self.color_default_action)

        self.color_by_entity_action = QAction("By Entity", self)
        self.color_by_entity_action.setCheckable(True)
        self.color_by_entity_action.triggered.connect(
            lambda: self.cdata.set_coloring_mode("entity")
        )
        coloring_group.addAction(self.color_by_entity_action)

        coloring_menu.addAction(self.color_default_action)
        coloring_menu.addAction(self.color_by_entity_action)

        legend_bar_menu = QMenu("Legend", self)
        legend_bar = QAction("Show", self)
        legend_bar.setCheckable(True)
        legend_bar.setChecked(False)
        legend_bar.triggered.connect(
            lambda checked: self.legend.show() if checked else self.legend.hide()
        )

        orientation_menu = QMenu("Orientation", self)
        vertical = QAction("Vertical", self)
        vertical.triggered.connect(lambda: self.legend.set_orientation("vertical"))
        horizontal = QAction("Horizontal", self)
        horizontal.triggered.connect(lambda: self.legend.set_orientation("horizontal"))

        orientation_menu.addAction(vertical)
        orientation_menu.addAction(horizontal)
        legend_bar_menu.addAction(legend_bar)
        legend_bar_menu.addMenu(orientation_menu)

        self.volume_action = QAction("Volume Viewer", self)
        self.volume_action.setCheckable(True)
        self.volume_action.setChecked(False)
        self.volume_action.triggered.connect(
            lambda checked: self.volume_dock.setVisible(checked)
        )

        self.trajectory_action = QAction("Trajectory Player", self)
        self.trajectory_action.setCheckable(True)
        self.trajectory_action.setChecked(False)
        self.trajectory_action.triggered.connect(
            lambda checked: self.trajectory_dock.setVisible(checked)
        )

        # Help menu
        show_keybinds_action = QAction("Keybinds", self)
        self.keybinds_dialog = KeybindsDialog(self)
        show_keybinds_action.triggered.connect(self.keybinds_dialog.show)
        show_keybinds_action.setShortcut("Ctrl+H")

        # Add actions to menus
        file_menu.addAction(add_file_action)
        file_menu.addMenu(self.recent_menu)

        file_menu.addSeparator()
        file_menu.addAction(new_session_action)
        file_menu.addAction(save_file_action)
        file_menu.addAction(close_file_action)

        file_menu.addSeparator()
        file_menu.addAction(screenshot_action)
        file_menu.addAction(clipboard_action)
        file_menu.addAction(clipboard_window_action)
        file_menu.addAction(animation_action)
        file_menu.addSeparator()
        file_menu.addAction(quit_action)

        show_scale_bar = QAction("Scale Bar", self)
        show_scale_bar.setCheckable(True)
        show_scale_bar.setChecked(False)
        show_scale_bar.triggered.connect(
            lambda checked: self.scale_bar.show() if checked else self.scale_bar.hide()
        )

        show_viewer_mode = QAction("Status Bar", self)
        show_viewer_mode.setCheckable(True)
        show_viewer_mode.setChecked(True)
        show_viewer_mode.triggered.connect(
            lambda checked: (
                self.status_indicator.show()
                if checked
                else self.status_indicator.hide()
            )
        )

        view_menu.addMenu(axes_menu)
        view_menu.addMenu(tilt_menu)
        view_menu.addMenu(legend_bar_menu)
        view_menu.addMenu(coloring_menu)

        view_menu.addSeparator()
        view_menu.addAction(show_scale_bar)
        view_menu.addAction(show_viewer_mode)

        view_menu.addSeparator()

        xy_action = QAction("XY-Plane", self)
        xy_action.setText("Top View (XY)\tz")
        xy_action.triggered.connect(lambda: self.simulate_key_press("z"))
        yz_action = QAction("YZ-Plane", self)
        yz_action.setText("Side View (YZ)\tx")
        yz_action.triggered.connect(lambda: self.simulate_key_press("x"))
        xz_action = QAction("XZ-Plane", self)
        xz_action.setText("Front View (XZ)\tc")
        xz_action.triggered.connect(lambda: self.simulate_key_press("c"))

        flip_action = QAction("Flip View lambda", self)
        flip_action.setText("Flip View Axis \tv")
        flip_action.triggered.connect(lambda: self.simulate_key_press("v"))

        view_menu.addAction(xy_action)
        view_menu.addAction(yz_action)
        view_menu.addAction(xz_action)
        view_menu.addAction(flip_action)
        view_menu.addSeparator()

        view_menu.addAction(self.volume_action)
        view_menu.addAction(self.trajectory_action)
        view_menu.addSeparator()

        bbox_menu = QMenu("Bounding Boxes", self)

        self.computed_bbox = QAction("Dataset Bounds", self)
        self.computed_bbox.setCheckable(True)
        self.computed_bbox.setChecked(False)
        self.computed_bbox.triggered.connect(
            lambda checked: self.bbox_manager.show_dataset_bounds(checked)
        )

        self.dataset_bbox = QAction("Session Bound", self)
        self.dataset_bbox.setCheckable(True)
        self.dataset_bbox.setChecked(False)
        self.dataset_bbox.triggered.connect(
            lambda checked: _handle_session_bounds(checked)
        )

        def _handle_session_bounds(checked):
            self.bbox_manager.show_session_bounds(checked)
            if self.cdata.shape is None:
                self.dataset_bbox.setChecked(False)

        show_all_objects = QAction("Show All Visible", self)
        show_all_objects.triggered.connect(self.bbox_manager.show_all_object_boxes)

        show_selected_objects = QAction("Show Selected", self)
        show_selected_objects.triggered.connect(self.bbox_manager.show_selected_boxes)

        hide_object_boxes = QAction("Hide All", self)
        hide_object_boxes.triggered.connect(self.bbox_manager.clear_object_boxes)

        bbox_menu.addAction(hide_object_boxes)
        bbox_menu.addAction(show_all_objects)
        bbox_menu.addAction(show_selected_objects)
        bbox_menu.addSeparator()
        bbox_menu.addAction(self.computed_bbox)
        bbox_menu.addAction(self.dataset_bbox)

        view_menu.addMenu(bbox_menu)
        view_menu.addSeparator()

        show_settings = QAction("Appearance", self)
        show_settings.triggered.connect(self.show_app_settings)
        preference_menu.addAction(show_settings)

        help_menu.addAction(show_keybinds_action)

        viewing_action = QAction("Viewing Mode\tEsc", self)
        viewing_action.triggered.connect(lambda: self.handle_escape_key())

        selection_action = QAction("Point Selection\tr", self)
        selection_action.triggered.connect(lambda: self.simulate_key_press("r"))

        expand_selection_action = QAction("Expand Selection\te", self)
        expand_selection_action.triggered.connect(lambda: self.simulate_key_press("e"))

        hide_unselected_action = QAction("Hide Unselected\th", self)
        hide_unselected_action.triggered.connect(lambda: self.simulate_key_press("h"))

        show_unselected_action = QAction("Show Unselected\tShift+H", self)
        show_unselected_action.triggered.connect(lambda: self.simulate_key_press("H"))

        picking_action = QAction("Pick Objects\tShift+E", self)
        picking_action.triggered.connect(lambda: self.simulate_key_press("E"))

        remove_action = QAction("Remove Selection\tDelete", self)
        remove_action.triggered.connect(lambda: self.simulate_key_press("\x7f"))

        merge_action = QAction("Merge Selection", self)
        merge_action.setText("Merge Selection\tm")
        merge_action.triggered.connect(lambda: self.simulate_key_press("m"))

        drawing_action = QAction("Free Hand Drawing", self)
        drawing_action.setText("Free Hand Drawing\ta")
        drawing_action.triggered.connect(lambda: self.simulate_key_press("a"))

        curve_action = QAction("Curve Drawing\tShift+A", self)
        curve_action.triggered.connect(lambda: self.simulate_key_press("A"))

        mesh_delete_action = QAction("Delete Mesh Triangles\tq", self)
        mesh_delete_action.triggered.connect(lambda: self.simulate_key_press("q"))

        mesh_add_action = QAction("Add Mesh Triangles\tShift+Q", self)
        mesh_add_action.triggered.connect(lambda: self.simulate_key_press("m"))

        interaction_target_menu = QMenu("Interaction Target", self)
        target_group = QActionGroup(self)
        target_group.setExclusive(True)
        self.cluster_target_action = QAction("Clusters\ts", self)
        self.cluster_target_action.setCheckable(True)
        self.cluster_target_action.setChecked(True)
        self.cluster_target_action.triggered.connect(
            lambda: self.simulate_key_press("s")
        )
        target_group.addAction(self.cluster_target_action)

        self.model_target_action = QAction("Models\ts", self)
        self.model_target_action.setCheckable(True)
        self.model_target_action.triggered.connect(lambda: self.simulate_key_press("s"))
        target_group.addAction(self.model_target_action)
        interaction_target_menu.addAction(self.cluster_target_action)
        interaction_target_menu.addAction(self.model_target_action)

        interact_menu.addAction(undo_action)
        interact_menu.addAction(viewing_action)
        interact_menu.addSeparator()

        interact_menu.addAction(selection_action)
        interact_menu.addAction(picking_action)
        interact_menu.addMenu(interaction_target_menu)
        interact_menu.addSeparator()

        interact_menu.addAction(merge_action)
        interact_menu.addAction(remove_action)
        interact_menu.addAction(expand_selection_action)
        interact_menu.addSeparator()

        interact_menu.addAction(hide_unselected_action)
        interact_menu.addAction(show_unselected_action)
        interact_menu.addSeparator()

        interact_menu.addAction(drawing_action)
        interact_menu.addAction(curve_action)
        interact_menu.addSeparator()

        interact_menu.addAction(mesh_add_action)
        interact_menu.addAction(mesh_delete_action)

    def toggle_selection_menu(self):
        """Update the menu radio buttons to reflect current selection target."""
        if self.model_target_action.isChecked():
            self.cluster_target_action.setChecked(True)
            self.status_indicator.update_status(target="Clusters")
        else:
            self.model_target_action.setChecked(True)
            self.status_indicator.update_status(target="Models")

    def simulate_key_press(self, key):
        self.vtk_widget.setFocus()

        key_code = (
            ord(key.upper())
            if len(key) == 1
            else getattr(Qt.Key, f"Key_{key}", ord(key))
        )

        key_press = QKeyEvent(
            QEvent.Type.KeyPress, key_code, Qt.KeyboardModifier.NoModifier, key
        )

        key_release = QKeyEvent(
            QEvent.Type.KeyRelease, key_code, Qt.KeyboardModifier.NoModifier, key
        )

        QApplication.postEvent(self.vtk_widget, key_press)
        QApplication.postEvent(self.vtk_widget, key_release)
        QApplication.processEvents()

    def _animate(self):
        from mosaic.animation.compose import AnimationComposerDialog

        dialog = AnimationComposerDialog(
            self.vtk_widget, self.volume_viewer, self.cdata
        )
        dock = QDockWidget("Animation Composer", self)
        dock.setFeatures(QDockWidget.DockWidgetFeature.NoDockWidgetFeatures)
        dock.setWidget(dialog)

        dock.setFeatures(
            QDockWidget.DockWidgetClosable
            | QDockWidget.DockWidgetFloatable
            | QDockWidget.DockWidgetMovable
        )

        dialog.accepted.connect(dock.close)
        dialog.rejected.connect(dock.close)

        self.volume_viewer.setMaximumHeight(100)
        self.volume_viewer.setMinimumHeight(50)
        self.addDockWidget(Qt.RightDockWidgetArea, dock)

        dock.raise_()
        dock.show()

    def _setup_volume_viewer(self):
        self.volume_dock = QDockWidget(parent=self)
        self.volume_dock.setFeatures(QDockWidget.DockWidgetFeature.NoDockWidgetFeatures)
        self.volume_dock.setTitleBarWidget(QWidget())

        self.volume_dock.setWidget(self.volume_viewer)
        self.addDockWidget(
            Qt.DockWidgetArea.BottomDockWidgetArea,
            self.volume_dock,
            Qt.Orientation.Vertical,
        )
        self.volume_dock.setVisible(False)

    def _setup_trajectory_player(self):
        self.trajectory_dock = QDockWidget(parent=self)
        self.trajectory_dock.setFeatures(
            QDockWidget.DockWidgetFeature.NoDockWidgetFeatures
        )
        self.trajectory_dock.setTitleBarWidget(QWidget())

        self.trajectory_player = TrajectoryPlayer(self.cdata)
        self.trajectory_dock.setWidget(self.trajectory_player)
        self.addDockWidget(
            Qt.DockWidgetArea.BottomDockWidgetArea,
            self.trajectory_dock,
            Qt.Orientation.Vertical,
        )
        self.trajectory_dock.setVisible(False)

    def show_app_settings(self):
        dialog = AppSettingsDialog(self)
        dialog.settingsChanged.connect(self.apply_render_settings)
        if dialog.exec() == 1:
            return self.apply_render_settings()

    def _load_session(self, file_path: str):
        self.close_session(show_warning=False)

        try:
            self.cdata.load_session(file_path)
        except ValueError as e:
            print(f"Error opening file: {e}")
            return -1

        self._add_file_to_recent(file_path)

        self.cdata.data.render()
        self.cdata.models.render()
        return self.reset_camera_view()

    def load_session(self):
        file_dialog = QFileDialog()
        file_path, _ = file_dialog.getOpenFileName(self, "Open Session")
        if not file_path:
            return -1
        return self._load_session(file_path)

    def close_session(self, show_warning: bool = True):

        def _show_close_session_warning() -> bool:
            msg_box = QMessageBox()
            msg_box.setIcon(QMessageBox.Icon.Warning)
            msg_box.setWindowTitle("Close Session")
            msg_box.setText("Close Session Warning")
            msg_box.setInformativeText(
                "Are you sure you want to close the current session? "
                "This action cannot be undone and all current work will be lost."
            )
            msg_box.setStandardButtons(
                QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel
            )
            msg_box.setDefaultButton(QMessageBox.StandardButton.Cancel)
            msg_box.button(QMessageBox.StandardButton.Ok).setText("Close")
            msg_box.button(QMessageBox.StandardButton.Cancel).setText("Keep")

            result = msg_box.exec()
            return result == QMessageBox.StandardButton.Ok

        if show_warning:
            if not _show_close_session_warning():
                return None

        self.renderer.RemoveAllViewProps()
        self.volume_viewer.close()

        self.dataset_bbox.setChecked(False)
        self.computed_bbox.setChecked(False)

        if self.scale_bar.visible:
            self.scale_bar.show()

        if self.status_indicator.visible:
            self.status_indicator.show()

        self.cdata.reset()
        self.cdata.data.render()
        self.cdata.models.render()

        self.set_camera_view("z")

    def _open_file(self, filename, parameters):
        from .formats import open_file

        offset = parameters.get("offset", 0)
        scale = parameters.get("scale", 1)
        sampling = parameters.get("sampling_rate", 1)

        try:
            container = open_file(filename)
        except Exception as e:
            if filename.endswith(".pickle"):
                raise ValueError("Use Load Session to open session files.")
            raise e

        base, _ = basename(filename).split(extsep, 1)
        use_index = len(container) > 1
        if len(container) > 1000:
            reply = QMessageBox.question(
                self,
                "Large number of objects detected",
                f"File '{basename(filename)}' contains {len(container):,} objects.\n\n"
                "This may indicate:\n"
                "- Raw EM data instead of segmentations or meshes\n"
                "- Incorrectly formatted file\n"
                "- You are dealing with a large dataset\n"
                "Processing will require considerable compute capabilities.\n\n"
                "Do you want to continue?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            )
            if reply != QMessageBox.StandardButton.Yes:
                return None

        for index, data in enumerate(container):
            # data.sampling is typically 1 apart from parser.read_volume
            scale_new = np.divide(scale, data.sampling)
            data.vertices = np.multiply(np.subtract(data.vertices, offset), scale_new)

            if data.vertices.shape[0] > 1e7:
                if not show_large_file_warning():
                    continue

            data_shape = np.divide(data.shape, data.sampling)

            name = base if not use_index else f"{index}_{base}"
            if data.faces is None:
                index = self.cdata.data.add(
                    points=data.vertices,
                    normals=data.normals,
                    sampling_rate=sampling,
                    quaternions=data.quaternions,
                    vertex_properties=data.vertex_properties,
                )
                mosaic_container = self.cdata._data

            else:
                from .meshing import to_open3d
                from .parametrization import TriangularMesh

                index = self.cdata._add_fit(
                    fit=TriangularMesh(to_open3d(data.vertices, data.faces)),
                    sampling_rate=sampling,
                    vertex_properties=data.vertex_properties,
                )
                mosaic_container = self.cdata._models

            geometry = mosaic_container.get(index)
            geometry._meta["name"] = name
            if parameters.get("render_as_surface", False):
                geometry.change_representation("surface")

            if "shape" not in mosaic_container.metadata:
                mosaic_container.metadata["shape"] = data_shape
            mosaic_container.metadata["shape"] = np.maximum(
                mosaic_container.metadata["shape"], data_shape
            )

    def _open_files(self, filenames: List[str]):
        dialog = ImportDataDialog(self)
        if isinstance(filenames, str):
            filenames = [filenames]

        dialog.set_files(filenames)
        if not dialog.exec():
            return -1

        file_parameters = dialog.get_all_parameters()
        with ProgressDialog(filenames, title="Reading Files", parent=None) as pbar:
            for filename in pbar:
                self._open_file(filename, file_parameters[filename])
                self._add_file_to_recent(filename)

        self.cdata.data.data_changed.emit()
        self.cdata.models.data_changed.emit()
        self.cdata.data.render()
        self.cdata.models.render()

        # Make sure loaded objects are visible in scene
        return self.set_camera_view("z")

    def open_files(self):
        filenames, _ = QFileDialog.getOpenFileNames(self, "Import Files")
        if not filenames:
            return -1

        return self._open_files(filenames)

    def save_session(self):
        file_dialog = QFileDialog()
        file_dialog.setDefaultSuffix("pickle")
        file_path, _ = file_dialog.getSaveFileName(
            self, "Save File", "", "Pickle Files (*.pickle)"
        )
        if not file_path:
            return -1

        # setDefaultSuffix seems appears across platforms
        if not file_path.lower().endswith(".pickle"):
            file_path += ".pickle"
        self.cdata.to_file(file_path)

    def update_recent_files_menu(self):
        files_to_show = list(dict.fromkeys(Settings.ui.recent_files))

        for i, file_path in enumerate(files_to_show):
            text = f"&{i + 1} {os.path.basename(file_path)}"
            self.recent_file_actions[i].setText(text)
            self.recent_file_actions[i].setData(file_path)
            self.recent_file_actions[i].setVisible(True)

        for j in range(len(files_to_show), Settings.ui.max_recent_files):
            self.recent_file_actions[j].setVisible(False)

        self.recent_menu.setEnabled(len(files_to_show) > 0)

    def _add_file_to_recent(self, file_path):
        if file_path in Settings.ui.recent_files:
            return None

        recent_files = [file_path] + list(Settings.ui.recent_files)
        while len(recent_files) > Settings.ui.max_recent_files:
            recent_files.pop()
        Settings.ui.recent_files = list(dict.fromkeys(recent_files))

        self.update_recent_files_menu()

    def _open_recent_file(self):
        action = self.sender()
        if not action:
            return None

        file_path = action.data()
        if not os.path.exists(file_path):
            QMessageBox.critical(self, "Error", f"{file_path} not found.")
            recent_files = list(Settings.ui.recent_files)
            try:
                recent_files.remove(file_path)
            except Exception:
                pass
            Settings.ui.recent_files = recent_files
            return self.update_recent_files_menu()

        if file_path.endswith(".pickle"):
            return self._load_session(file_path)
        return self._open_files([file_path])

    def _check_for_updates(self):
        from .dialogs import UpdateChecker, UpdateDialog
        from .__version__ import __version__

        def _show_update_dialog(latest_version, release_notes):
            if Settings.ui.skipped_version == latest_version:
                return None
            dialog = UpdateDialog(
                __version__, latest_version, release_notes, parent=self
            )
            dialog.exec()

        # We assign the thread to keep it alive
        self.update_checker = UpdateChecker(__version__)
        self.update_checker.update_available.connect(_show_update_dialog)
        self.update_checker.start()


def show_large_file_warning() -> bool:
    if Settings.warnings.suppress_large_file_warning:
        return True

    msg_box = QMessageBox()
    msg_box.setIcon(QMessageBox.Icon.Warning)
    msg_box.setWindowTitle("Large File Detected")

    msg_box.setText(
        "Large File Warning\n\n"
        "We found one or more files exceeding 10 million points."
    )

    msg_box.setInformativeText(
        "Please make sure this is a segmentation and not raw data. "
        "If you are on a laptop without dedicated GPU, consider reducing the number "
        "of points for a smooth experience using Segmentation > Downsample "
        "or process in batches."
    )

    msg_box.setStandardButtons(
        QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel
    )
    msg_box.setDefaultButton(QMessageBox.StandardButton.Ok)
    msg_box.button(QMessageBox.StandardButton.Ok).setText("Accept")
    msg_box.button(QMessageBox.StandardButton.Cancel).setText("Skip File")

    help_button = msg_box.addButton("Help", QMessageBox.ButtonRole.HelpRole)

    def open_documentation():
        import webbrowser

        webbrowser.open(
            "https://kosinskilab.github.io/mosaic/tutorial/reference/troubleshooting.html#performance-issues"
        )

    help_button.clicked.connect(open_documentation)

    checkbox = QCheckBox("Don't show this warning again")
    msg_box.setCheckBox(checkbox)

    result = msg_box.exec()

    if checkbox.isChecked():
        Settings.warnings.suppress_large_file_warning = True

    return result == QMessageBox.StandardButton.Ok



from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QMessageBox,
    QDockWidget,
    QApplication,
    QMainWindow,
)


def create_or_toggle_dock(
    instance, dock_attr_name, dialog_widget, dock_area=Qt.RightDockWidgetArea
):
    """
    Helper method to create or toggle a docked dialog.

    Parameters
    ----------
    dock_attr_name : str
        The attribute name to store the dock widget (e.g., 'histogram_dock')
    dialog_widget : QWidget
        The dialog widget to display in the dock
    dock_area : Qt.DockWidgetArea, optional
        Where to dock the widget, default is RightDockWidgetArea
    """

    def _exit():
        dock = getattr(instance, dock_attr_name, None)
        if dock:
            if widget := dock.widget():
                widget.close()
            dock.close()
            dock.deleteLater()
        setattr(instance, dock_attr_name, None)

    if getattr(instance, dock_attr_name, None) is not None:
        return _exit()

    dock = QDockWidget()
    dock.setFeatures(
        QDockWidget.DockWidgetClosable
        | QDockWidget.DockWidgetFloatable
        | QDockWidget.DockWidgetMovable
    )
    dock.setWidget(dialog_widget)

    # Handle cleanup when dock is closed via X button
    dock.visibilityChanged.connect(lambda visible: _exit() if not visible else None)

    if hasattr(dialog_widget, "accepted"):
        dialog_widget.accepted.connect(_exit)
    if hasattr(dialog_widget, "rejected"):
        dialog_widget.rejected.connect(_exit)

    main_window = None
    for widget in QApplication.instance().topLevelWidgets():
        if isinstance(widget, QMainWindow):
            main_window = widget
            break

    if main_window is None:
        QMessageBox.warning(
            instance, "Warning", "Could not determine application main window."
        )
        return dialog_widget.show()

    main_window.addDockWidget(dock_area, dock)
    setattr(instance, dock_attr_name, dock)
    dock.show()
    dock.raise_()



import numpy as np
from qtpy.QtCore import Qt, QTimer
from qtpy.QtWidgets import (
    QDialog,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QComboBox,
    QListWidget,
    QGroupBox,
    QCheckBox,
    QSpinBox,
    QPushButton,
    QFormLayout,
    QWidget,
    QMessageBox,
    QTabWidget,
    QTableWidget,
    QHeaderView,
    QTableWidgetItem,
    QFileDialog,
)
import pyqtgraph as pg
import qtawesome as qta

from ..widgets.settings import get_widget_value, set_widget_value
from ..stylesheets import QPushButton_style, QScrollArea_style
from ..widgets import ContainerListWidget, StyledListWidgetItem, ColorPreviewWidget


def _populate_list(geometries):
    target_list = ContainerListWidget(border=False)
    target_list.setSelectionMode(QListWidget.SelectionMode.SingleSelection)

    for name, obj in geometries:
        item = StyledListWidgetItem(name, obj.visible, obj._meta.get("info"))
        item.setData(Qt.ItemDataRole.UserRole, obj)
        target_list.addItem(item)
    return target_list


class PropertyAnalysisDialog(QDialog):
    def __init__(self, cdata, legend=None, parent=None):
        super().__init__(parent)
        self.cdata = cdata
        self.properties = {}
        self.property_parameters = {}

        self.setWindowTitle("Property Analysis")

        self.legend = legend
        self.color_preview = ColorPreviewWidget()
        self.setWindowFlags(Qt.WindowType.Window)

        self._setup_ui()
        self._setup_styling()

        self.cdata.data.vtk_pre_render.connect(self._on_render_update)
        self.cdata.models.vtk_pre_render.connect(self._on_render_update)

    def _on_render_update(self):
        """Re-apply properties when models are re-rendered"""
        self.cdata.data.blockSignals(True)
        self.cdata.models.blockSignals(True)
        try:
            self._update_property_list()

            # Provoke cache miss for tracking inclusions on trajectories
            self.property_parameters.clear()
            self._preview(render=False, suppress_warning=True)
            self._update_plot()
            self._update_statistics()
        except Exception:
            # Things like select at least one object
            pass
        finally:
            self.cdata.data.blockSignals(False)
            self.cdata.models.blockSignals(False)

    def closeEvent(self, event):
        """Disconnect when dialog closes"""
        try:
            self.cdata.data.vtk_pre_render.disconnect(self._on_render_update)
            self.cdata.models.vtk_pre_render.disconnect(self._on_render_update)
        except Exception:
            pass
        super().closeEvent(event)

    def _setup_ui(self):
        main_layout = QVBoxLayout(self)

        self.tabs_container = QWidget()
        tabs_layout = QVBoxLayout(self.tabs_container)
        tabs_layout.setContentsMargins(0, 0, 0, 0)

        self.tabs_widget = QTabWidget()
        self.visualization_tab = QWidget()
        self.analysis_tab = QWidget()
        self.statistics_tab = QWidget()

        self._setup_visualization_tab()
        self._setup_analysis_tab()
        self._setup_statistics_tab()

        # Add tabs with icons
        self.tabs_widget.addTab(
            self.visualization_tab, qta.icon("mdi.brush", color="#4f46e5"), "Visualize"
        )
        self.tabs_widget.addTab(
            self.analysis_tab,
            qta.icon("mdi.chart-bell-curve", color="#4f46e5"),
            "Distribution",
        )
        self.tabs_widget.addTab(
            self.statistics_tab,
            qta.icon("mdi.chart-bar", color="#4f46e5"),
            "Statistics",
        )
        self.tabs_widget.currentChanged.connect(self._update_tab)
        main_layout.addWidget(self.tabs_widget)

    def _setup_visualization_tab(self):
        from ..icons import dialog_accept_icon

        layout = QVBoxLayout(self.visualization_tab)

        property_group = QGroupBox("Property")
        property_layout = QVBoxLayout()

        category_layout = QHBoxLayout()
        category_layout.addWidget(QLabel("Category:"))
        self.category_combo = QComboBox()
        self.category_combo.addItems(
            ["Distance", "Surface", "Geometric", "Projection", "Custom"]
        )
        self.category_combo.currentTextChanged.connect(self._update_property_list)
        category_layout.addWidget(self.category_combo)

        category_layout.addSpacing(15)
        category_layout.addWidget(QLabel("Property:"))
        self.property_combo = QComboBox()
        self.property_combo.currentTextChanged.connect(self._update_options)
        category_layout.addWidget(self.property_combo, 1)
        property_layout.addLayout(category_layout)

        # Property-specific options container
        self.property_options_container = QWidget()
        self.property_options_layout = QFormLayout(self.property_options_container)
        self.property_options_layout.setContentsMargins(0, 10, 0, 0)
        property_layout.addWidget(self.property_options_container)

        property_group.setLayout(property_layout)
        layout.addWidget(property_group)

        options_group = QGroupBox("Visualization Options")
        options_group.setFixedHeight(150)
        options_layout = QVBoxLayout(options_group)

        colormap_layout = QHBoxLayout()
        colormap_layout.addWidget(QLabel("Color Map:"))

        self.colormap_combo = QComboBox()
        self.colormap_combo.addItems(self.color_preview.colormaps.copy())
        self.colormap_combo.currentTextChanged.connect(self._update_colormap_preview)
        colormap_layout.addWidget(self.colormap_combo)

        checkbox_layout = QHBoxLayout()
        self.normalize_checkbox = QCheckBox("Normalize per Object")
        checkbox_layout.addWidget(self.normalize_checkbox)
        self.quantile_checkbox = QCheckBox("Compute Quantiles")
        checkbox_layout.addWidget(self.quantile_checkbox)
        self.invert_checkbox = QCheckBox("Invert Colors")
        self.invert_checkbox.stateChanged.connect(self._update_colormap_preview)
        checkbox_layout.addWidget(self.invert_checkbox)

        options_layout.addLayout(colormap_layout)
        options_layout.addWidget(self.color_preview)
        options_layout.addLayout(checkbox_layout)
        layout.addWidget(options_group)

        # Dialog Control Buttons
        button_layout = QHBoxLayout()
        refresh_btn = QPushButton("Compute")
        refresh_btn.setIcon(qta.icon("mdi.monitor", color="#4f46e5"))
        refresh_btn.clicked.connect(self._preview)
        button_layout.addWidget(refresh_btn)
        button_layout.addStretch()

        show_dist_btn = QPushButton("Show Distribution")
        show_dist_btn.setIcon(qta.icon("mdi.chart-bell-curve", color="#4f46e5"))
        show_dist_btn.clicked.connect(lambda: self.tabs_widget.setCurrentIndex(1))
        button_layout.addWidget(show_dist_btn)

        export_btn = QPushButton("Export Data")
        export_btn.setIcon(qta.icon("mdi.download", color="#4f46e5"))
        export_btn.clicked.connect(self._export_data)
        button_layout.addWidget(export_btn)

        apply_btn = QPushButton("Done")
        apply_btn.setIcon(dialog_accept_icon)
        apply_btn.clicked.connect(self.close)
        button_layout.addWidget(apply_btn)
        layout.addLayout(button_layout)

        self._update_property_list("Distance")
        self._update_colormap_preview()

    def _setup_analysis_tab(self):
        from ..icons import dialog_accept_icon

        layout = QVBoxLayout(self.analysis_tab)

        plot_group = QGroupBox("Distribution")
        plot_layout = QVBoxLayout(plot_group)

        header_layout = QHBoxLayout()
        header_layout.addStretch()

        plot_type_layout = QHBoxLayout()
        self.plot_types = ["Histogram", "Density", "Line"]
        self.current_plot_type = "Density"
        self.bar_btn = QPushButton()
        self.bar_btn.setIcon(qta.icon("mdi.chart-histogram", color="#4f46e5"))
        self.bar_btn.setToolTip("Histogram")
        self.bar_btn.setFixedSize(28, 28)
        self.bar_btn.clicked.connect(lambda: self._set_plot_type("Histogram"))

        self.density_btn = QPushButton()
        self.density_btn.setIcon(qta.icon("mdi.chart-bell-curve", color="#4f46e5"))
        self.density_btn.setFixedSize(28, 28)
        self.density_btn.clicked.connect(lambda: self._set_plot_type("Density"))

        self.line_btn = QPushButton()
        self.line_btn.setIcon(qta.icon("mdi.chart-line", color="#4f46e5"))
        self.line_btn.setToolTip("Line Chart")
        self.line_btn.setFixedSize(28, 28)
        self.line_btn.clicked.connect(lambda: self._set_plot_type("Line"))
        plot_type_layout.addWidget(self.bar_btn)
        plot_type_layout.addWidget(self.density_btn)
        plot_type_layout.addWidget(self.line_btn)
        header_layout.addLayout(plot_type_layout)
        plot_layout.addLayout(header_layout)

        self.plot_widget = pg.GraphicsLayoutWidget(self)
        self.plot_widget.setBackground(None)
        plot_layout.addWidget(self.plot_widget)

        options_group = QGroupBox("Visualization Options")
        options_group.setFixedHeight(150)
        options_layout = QVBoxLayout(options_group)

        strat_layout = QHBoxLayout()
        self.plot_title = QLabel("Stratification")
        self.plot_mode_combo = QComboBox()
        self.plot_mode_combo.addItems(["Combined", "Separate"])
        self.plot_mode_combo.currentTextChanged.connect(self._update_plot)
        strat_layout.addWidget(self.plot_title)
        strat_layout.addWidget(self.plot_mode_combo)
        options_layout.addLayout(strat_layout)

        alpha_layout = QHBoxLayout()
        alpha_layout.addWidget(QLabel("Alpha:"))
        self.alpha_slider = QSpinBox()
        self.alpha_slider.setRange(0, 255)
        self.alpha_slider.setValue(128)
        self.alpha_slider.valueChanged.connect(self._update_plot)
        alpha_layout.addWidget(self.alpha_slider)
        options_layout.addLayout(alpha_layout)

        colormap_layout = QHBoxLayout()
        colormap_layout.addWidget(QLabel("Color Palette:"))

        self.analysis_colormap_combo = QComboBox()
        self.analysis_colormap_combo.addItems(self.color_preview.colormaps)
        self.analysis_colormap_combo.currentTextChanged.connect(self._update_plot)
        colormap_layout.addWidget(self.analysis_colormap_combo)
        options_layout.addLayout(colormap_layout)

        layout.addWidget(plot_group)
        layout.addWidget(options_group)

        # Dialog Control Buttons
        button_layout = QHBoxLayout()
        refresh_btn = QPushButton("Refresh")
        refresh_btn.setIcon(qta.icon("mdi.refresh", color="#4f46e5"))
        refresh_btn.clicked.connect(self._update_plot)
        button_layout.addWidget(refresh_btn)
        button_layout.addStretch()

        save_plot_btn = QPushButton("Save Plot")
        save_plot_btn.setIcon(qta.icon("mdi.content-save", color="#4f46e5"))
        save_plot_btn.clicked.connect(self._export_plot)
        button_layout.addWidget(save_plot_btn)

        export_btn = QPushButton("Export Data")
        export_btn.setIcon(qta.icon("mdi.download", color="#4f46e5"))
        export_btn.clicked.connect(self._export_data)
        button_layout.addWidget(export_btn)

        apply_btn = QPushButton("Done")
        apply_btn.setIcon(dialog_accept_icon)
        apply_btn.clicked.connect(self.close)
        button_layout.addWidget(apply_btn)
        layout.addLayout(button_layout)

    def _setup_statistics_tab(self):
        from ..icons import dialog_accept_icon

        layout = QVBoxLayout(self.statistics_tab)

        stats_group = QGroupBox("Statistics")
        stats_layout = QVBoxLayout(stats_group)
        self.stats_table = QTableWidget()
        self.stats_table.setColumnCount(5)
        self.stats_table.setHorizontalHeaderLabels(
            ["Object", "Min", "Max", "Mean", "Std Dev"]
        )
        self.stats_table.horizontalHeader().setSectionResizeMode(
            QHeaderView.ResizeMode.Stretch
        )
        stats_layout.addWidget(self.stats_table)

        layout.addWidget(stats_group)

        button_layout = QHBoxLayout()
        refresh_btn = QPushButton("Refresh")
        refresh_btn.setIcon(qta.icon("mdi.refresh", color="#4f46e5"))
        refresh_btn.clicked.connect(self._update_statistics)
        button_layout.addWidget(refresh_btn)
        button_layout.addStretch()

        save_plot_btn = QPushButton("Save Plot")
        save_plot_btn.setIcon(qta.icon("mdi.content-save", color="#4f46e5"))
        save_plot_btn.clicked.connect(self._export_plot)
        button_layout.addWidget(save_plot_btn)

        export_btn = QPushButton("Export Data")
        export_btn.setIcon(qta.icon("mdi.download", color="#4f46e5"))
        export_btn.clicked.connect(self._export_data)
        button_layout.addWidget(export_btn)

        apply_btn = QPushButton("Done")
        apply_btn.setIcon(dialog_accept_icon)
        apply_btn.clicked.connect(self.close)
        button_layout.addWidget(apply_btn)
        layout.addLayout(button_layout)

    def _update_property_list(self, category: str = None):
        if category is None:
            category = self.category_combo.currentText()

        properties = {
            "Distance": [
                "To Camera",
                "To Cluster",
                "To Model",
                "To Self",
            ],
            "Surface": [
                "Curvature",
                "Edge Length",
                "Surface Area",
                "Triangle Area",
                "Volume",
                "Triangle Volume",
                "Number of Vertices",
                "Number of Triangles",
            ],
            "Projection": [
                "Projected Curvature",
                "Geodesic Distance",
            ],
            "Geometric": [
                "Identity",
                "Width (X-axis)",
                "Depth (Y-axis)",
                "Height (Z-axis)",
                "Number of Points",
            ],
            "Custom": ["Vertex Properties"],
        }
        self.property_map = {
            # Distance
            "To Camera": "distance",
            "To Cluster": "distance",
            "To Model": "distance",
            "To Self": "distance",
            # Surface
            "Curvature": "mesh_curvature",
            "Edge Length": "mesh_edge_length",
            "Surface Area": "mesh_surface_area",
            "Triangle Area": "mesh_triangle_area",
            "Volume": "mesh_volume",
            "Triangle Volume": "mesh_triangle_volume",
            "Number of Vertices": "mesh_vertices",
            "Number of Triangles": "mesh_triangles",
            # Geometric
            "Identity": "identity",
            "Width (X-axis)": "width",
            "Depth (Y-axis)": "depth",
            "Height (Z-axis)": "height",
            "Number of Points": "n_points",
            # Projection
            "Projected Curvature": "projected_curvature",
            "Geodesic Distance": "geodesic_distance",
            # Custom
            "Vertex Properties": "vertex_property",
        }
        previous_text = self.property_combo.currentText()

        self.property_combo.blockSignals(True)
        self.property_combo.clear()
        self.property_combo.addItems(properties.get(category, []))
        if previous_text is not None:
            index = self.property_combo.findText(previous_text)
            if index >= 0:
                self.property_combo.setCurrentIndex(index)

        if self.property_combo.count() > 0:
            self._update_options(self.property_combo.currentText())
        self.property_combo.blockSignals(False)

    def _update_options(self, property_name: str = None):
        if property_name is None:
            property_name = self.property_combo.currentText()

        previous_parameters = {}
        if hasattr(self, "option_widgets"):
            previous_parameters = {
                k: get_widget_value(w)
                for k, w in self.option_widgets.items()
                if not isinstance(w, (QListWidget, ContainerListWidget))
            }

        while self.property_options_layout.rowCount() > 0:
            self.property_options_layout.removeRow(0)

        self.option_widgets = {}
        if property_name == "Vertex Properties":

            geometries = self._get_all_geometries()

            # For now use all instead of shared vertex properties
            properties = set()
            for geometry in geometries:
                if geometry.vertex_properties is None:
                    continue
                properties |= set(geometry.vertex_properties.properties)

            if len(properties) == 0:
                return self.property_combo.clear()

            options = QComboBox()
            options.addItems(sorted(list(properties)))

            self.property_options_layout.addRow("Type:", options)
            self.option_widgets["name"] = options

        elif property_name == "Curvature":
            curvature_combobox = QComboBox()
            curvature_combobox.addItems(["Gaussian", "Mean"])

            radius_spinbox = QSpinBox()
            radius_spinbox.setRange(1, 20)
            radius_spinbox.setValue(5)
            self.property_options_layout.addRow("Method:", curvature_combobox)
            self.property_options_layout.addRow("Radius:", radius_spinbox)

            self.option_widgets["curvature"] = curvature_combobox
            self.option_widgets["radius"] = radius_spinbox

        elif property_name == "Projected Curvature":
            # Target mesh selection
            target_group = QGroupBox("Target Mesh")
            target_layout = QVBoxLayout(target_group)

            target_list = _populate_list(
                self.cdata.format_datalist("models", mesh_only=True)
            )
            target_layout.addWidget(target_list)

            options_layout = QFormLayout()

            curvature_combobox = QComboBox()
            curvature_combobox.addItems(["Gaussian", "Mean"])
            options_layout.addRow("Method:", curvature_combobox)

            radius_spinbox = QSpinBox()
            radius_spinbox.setRange(1, 20)
            radius_spinbox.setValue(5)
            options_layout.addRow("Radius:", radius_spinbox)

            target_layout.addLayout(options_layout)

            self.property_options_layout.addRow(target_group)
            self.option_widgets["queries"] = target_list
            self.option_widgets["curvature"] = curvature_combobox
            self.option_widgets["radius"] = radius_spinbox

        elif property_name == "Geodesic Distance":
            target_group = QGroupBox("Target Mesh")
            target_layout = QVBoxLayout(target_group)

            target_list = _populate_list(
                self.cdata.format_datalist("models", mesh_only=True)
            )
            target_layout.addWidget(target_list)

            neighbor_layout = QHBoxLayout()
            neighbor_label = QLabel("k-Nearest Neighbors:")
            knn_layout = QHBoxLayout()

            neighbor_start = QSpinBox()
            neighbor_start.setRange(1, 255)
            neighbor_start.setValue(1)

            neighbor_to_label = QLabel("to")
            neighbor_to_label.setAlignment(Qt.AlignmentFlag.AlignCenter)

            neighbor_end = QSpinBox()
            neighbor_end.setRange(1, 255)
            neighbor_end.setValue(1)

            knn_layout.addWidget(neighbor_start)
            knn_layout.addWidget(neighbor_to_label)
            knn_layout.addWidget(neighbor_end)
            neighbor_layout.addWidget(neighbor_label)
            neighbor_layout.addLayout(knn_layout)
            target_layout.addLayout(neighbor_layout)

            self.property_options_layout.addRow(target_group)
            self.option_widgets["queries"] = target_list
            self.option_widgets["k_start"] = neighbor_start
            self.option_widgets["k"] = neighbor_end

        elif property_name in ("To Cluster", "To Self"):

            target_group = QGroupBox("Options")
            target_layout = QVBoxLayout(target_group)

            if property_name == "To Cluster":

                target_list = _populate_list(self.cdata.format_datalist("data"))
                target_layout.addWidget(target_list)

                # Checkboxes
                all_targets_checkbox = QCheckBox("Compare to All")
                include_self_checkbox = QCheckBox("Include Within-Cluster Distance")
                all_targets_checkbox.stateChanged.connect(
                    lambda state: self.toggle_all_targets(state, target_list)
                )

                checkbox_layout = QHBoxLayout()
                checkbox_layout.addWidget(all_targets_checkbox)
                checkbox_layout.addWidget(include_self_checkbox)
                target_layout.addLayout(checkbox_layout)

                self.option_widgets["queries"] = target_list
                self.option_widgets["include_self"] = include_self_checkbox
                self.option_widgets["compare_to_all"] = all_targets_checkbox

            if property_name == "To Self":
                self_checkbox = QCheckBox()
                self_checkbox.setChecked(True)
                self.option_widgets["only_self"] = self_checkbox

            # KNN range
            neighbor_layout = QHBoxLayout()
            neighbor_label = QLabel("k-Nearest Neighbors:")
            knn_layout = QHBoxLayout()

            neighbor_start = QSpinBox()
            neighbor_start.setRange(1, 255)
            neighbor_start.setValue(1)

            neighbor_to_label = QLabel("to")
            neighbor_to_label.setAlignment(Qt.AlignmentFlag.AlignCenter)

            neighbor_end = QSpinBox()
            neighbor_end.setRange(1, 255)
            neighbor_end.setValue(1)

            neighbor_start.valueChanged.connect(lambda x: neighbor_end.setRange(x, 255))

            knn_layout.addWidget(neighbor_start)
            knn_layout.addWidget(neighbor_to_label)
            knn_layout.addWidget(neighbor_end)
            neighbor_layout.addWidget(neighbor_label)
            neighbor_layout.addLayout(knn_layout)
            target_layout.addLayout(neighbor_layout)

            self.property_options_layout.addRow(target_group)

            self.option_widgets["k_start"] = neighbor_start
            self.option_widgets["k"] = neighbor_end

        elif property_name == "To Model":
            target_group = QGroupBox("Target Models")
            target_layout = QVBoxLayout(target_group)

            target_list = _populate_list(self.cdata.format_datalist("models"))
            target_layout.addWidget(target_list)

            all_targets_checkbox = QCheckBox("Compare to All")
            checkbox_layout = QHBoxLayout()
            checkbox_layout.addWidget(all_targets_checkbox)
            target_layout.addLayout(checkbox_layout)

            all_targets_checkbox.stateChanged.connect(
                lambda state: self.toggle_all_targets(state, target_list)
            )

            self.property_options_layout.addRow(target_group)
            self.option_widgets["queries"] = target_list
            self.option_widgets["compare_to_all"] = all_targets_checkbox

        # Avoid re-entering parameters over and over
        for k in self.option_widgets.keys():
            if k not in previous_parameters:
                continue
            set_widget_value(self.option_widgets[k], previous_parameters[k])

    def toggle_all_targets(self, state, target_list):
        target_list.setEnabled(not bool(state))

        if bool(state):
            for i in range(target_list.count()):
                item = target_list.item(i)
                item.setSelected(True)
        else:
            target_list.clearSelection()

    def _update_colormap_preview(self):
        cmap = self.colormap_combo.currentText()
        reverse = self.invert_checkbox.isChecked()
        self.color_preview.set_colormap(cmap, reverse)

    def _get_selected_geometries(self):
        return [x[1] for x in self._get_selection()]

    def _get_all_geometries(self):
        return [x[1] for x in self._get_selection(selected=False)]

    def _get_selection(self, selected: bool = True):
        return [
            *self.cdata.format_datalist("data", selected=selected),
            *self.cdata.format_datalist("models", selected=selected),
        ]

    def _compute_properties(self):
        from ..properties import GeometryProperties

        options = {}
        for k, widget in self.option_widgets.items():
            if isinstance(widget, (QListWidget, ContainerListWidget)):
                value = [
                    item.data(Qt.ItemDataRole.UserRole)
                    for item in widget.selectedItems()
                ]
            else:
                value = get_widget_value(widget)
            options[k] = value

        # Assuming identical parameters, which geometric properties need computation
        missing_geometries = []
        geometries = self._get_selected_geometries()
        for geometry in geometries:
            geometry_id = id(geometry)

            # Newly selected object
            value = self.properties.get(geometry_id)
            if value is None:
                missing_geometries.append(geometry)
                continue

            # In case the object was modified during the dialog lifetime
            # TODO: Add listener to data_changed to track changes in aggregated metrics
            if hasattr(value, "size"):
                if value.size != geometry.points.shape[0]:
                    missing_geometries.append(geometry)

        # This could be explicity by making the compute / refresh buttons not
        # clickable if the properties are invalid
        property_name = self.property_map.get(self.property_combo.currentText())
        if property_name is None:
            return None

        options["property_name"] = property_name
        if self.property_combo.currentText() == "To Camera":
            vtk_widget = self.cdata.data.vtk_widget
            renderer = vtk_widget.GetRenderWindow().GetRenderers().GetFirstRenderer()
            options["queries"] = np.array(
                renderer.GetActiveCamera().GetPosition()
            ).reshape(1, -1)

        # Recompute all properties if parameters changed
        cache_miss = len(options) != len(self.property_parameters)
        for key, value in options.items():
            if key not in self.property_parameters:
                cache_miss = True
            other_value = self.property_parameters.get(key)

            try:
                if isinstance(value, np.ndarray) or isinstance(other_value, np.ndarray):
                    cache_miss = not np.allclose(value, other_value)

                cache_miss = value != other_value
                if not isinstance(cache_miss, bool):
                    cache_miss = all(cache_miss)
            except Exception:
                cache_miss = True

            if cache_miss:
                missing_geometries = geometries
                self.properties.clear()
                break

        self.property_parameters = options
        if options["property_name"] == "identity":
            self.properties = {id(x): i for i, x in enumerate(geometries)}
            return None

        try:
            self.properties.update(
                {
                    id(x): GeometryProperties.compute(geometry=x, **options)
                    for i, x in enumerate(missing_geometries)
                }
            )
        except Exception as e:
            QMessageBox.warning(self, "Error", str(e))
            self.properties.clear()

    def _preview(self, render: bool = True, suppress_warning: bool = False):
        from ..utils import cmap_to_vtkctf

        geometries = self._get_selected_geometries()
        if not geometries:
            if not suppress_warning:
                QMessageBox.warning(
                    self, "No Selection", "Please select at least one object."
                )
            return None

        self._compute_properties()
        colormap = self.colormap_combo.currentText()
        if self.invert_checkbox.isChecked():
            colormap += "_r"

        properties = self.properties
        if self.normalize_checkbox.isChecked():
            properties = {
                k: (v - v.min()) / (v.max() - v.min()) for k, v in properties.items()
            }

        if self.quantile_checkbox.isChecked():
            all_curvatures = np.concatenate([v.flatten() for v in properties.values()])
            valid_curvatures = all_curvatures[~np.isnan(all_curvatures)]
            n_bins = min(valid_curvatures.size // 10, 100)
            bins = np.percentile(valid_curvatures, np.linspace(0, 100, n_bins + 1))
            properties = {k: np.digitize(v, bins) - 1 for k, v in properties.items()}

        values = [x for x in properties.values() if x is not None]
        if len(values) == 0:
            return None

        max_value = np.max([np.max(x) for x in values])
        min_value = np.min([np.min(x) for x in values])
        lut, lut_range = cmap_to_vtkctf(colormap, max_value, min_value=min_value)
        for geometry in geometries:
            metric = properties.get(id(geometry))
            if metric is None:
                continue
            geometry.set_scalars(metric, lut, lut_range)

        self.legend.set_lookup_table(lut, self.property_combo.currentText())

        if render:
            self.cdata.data.render_vtk()
            self.cdata.models.render_vtk()

    def _update_tab(self):
        current_tab_index = self.tabs_widget.currentIndex()

        self.plot_widget.clear()
        QTimer.singleShot(
            100,
            lambda: (
                self._update_plot()
                if current_tab_index == 1
                else self._update_statistics() if current_tab_index == 2 else None
            ),
        )

    def _update_statistics(self):
        selected_items = self._get_selection()
        self.stats_table.setRowCount(len(selected_items))

        row_count, n_decimals = 0, 6
        for index, (item_text, obj) in enumerate(selected_items):

            value = self.properties.get(id(obj))
            if value is None:
                continue

            row_count += 1
            item = QTableWidgetItem(item_text)
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.stats_table.setItem(index, 0, item)

            item = QTableWidgetItem(str(np.round(np.min(value), n_decimals)))
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.stats_table.setItem(index, 1, item)

            item = QTableWidgetItem(str(np.round(np.max(value), n_decimals)))
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.stats_table.setItem(index, 2, item)

            item = QTableWidgetItem(str(np.round(np.mean(value), n_decimals)))
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.stats_table.setItem(index, 3, item)

            item = QTableWidgetItem(str(np.round(np.std(value), n_decimals)))
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.stats_table.setItem(index, 4, item)
        self.stats_table.setRowCount(row_count)

    def _set_plot_type(self, plot_type):
        self.current_plot_type = plot_type
        self._update_plot()

    def _update_plot(self):
        """Update the plot based on the current property and selected objects"""
        if self.tabs_widget.currentIndex() != 1:
            return None

        selected_items = self._get_selection()
        if not selected_items or not hasattr(self, "properties"):
            return None

        plot_type = getattr(self, "current_plot_type", "Density")
        plot_mode = getattr(self, "plot_mode_combo", lambda: "Combined").currentText()
        alpha = getattr(self, "alpha_slider", lambda: 150).value()
        colormap = getattr(
            self, "analysis_colormap_combo", lambda: "viridis"
        ).currentText()
        colors = self.color_preview.generate_gradient(colormap, len(selected_items))
        colors = [pg.mkColor(c.red(), c.green(), c.blue(), alpha) for c in colors]

        data_series = []
        all_values = []
        for i, (item_text, obj) in enumerate(selected_items):
            obj_id = id(obj)
            values = self.properties.get(obj_id)
            if values is not None:
                all_values.append(values)
                data_series.append((item_text, obj, values, colors[i % len(colors)]))

        if not data_series:
            return None

        try:
            self.plot_widget.setUpdatesEnabled(False)

            self.plot_widget.clear()
            all_scalar = not isinstance(all_values[0], np.ndarray)
            if all_scalar:
                all_values = np.asarray(all_values)
                self._create_categorical_plot(data_series, all_values, plot_type)
            else:
                self._create_plot(data_series, all_values, plot_mode, plot_type)
        finally:
            self.plot_widget.setUpdatesEnabled(True)

    def _create_categorical_plot(self, data_series, values, plot_type):
        """Create a categorical plot with names on x-axis for single values"""
        property_name = self.property_combo.currentText()

        plot = self.plot_widget.addPlot()
        plot.setLabel("left", property_name)

        ax = plot.getAxis("bottom")
        names = [name for name, _, _, _ in data_series]
        colors = [color for _, _, _, color in data_series]
        ax.setTicks([[(i, name) for i, name in enumerate(names)]])

        if plot_type == "Histogram":
            for i in range(len(data_series)):
                bar = pg.BarGraphItem(
                    x=[i],
                    height=[values[i]],
                    width=0.7,
                    brush=colors[i],
                    pen=pg.mkPen("k", width=1),
                )
                plot.addItem(bar)
        else:
            scatter = pg.ScatterPlotItem()
            min_val, max_val = min(values), max(values)
            range_val = max_val - min_val if max_val > min_val else 1

            for i, (name, _, value, color) in enumerate(data_series):
                size = 10 + 40 * (values[i] - min_val) / range_val
                scatter.addPoints(
                    x=[i],
                    y=[values[i]],
                    size=size,
                    brush=color,
                    pen=pg.mkPen("k", width=1),
                    name=name,
                )
            plot.addItem(scatter)
        plot.addLegend(offset=(-10, 10))

    def _create_plot(self, data_series, all_values, plot_mode, plot_type):
        """Create either histogram, density or line plot based on plot_type"""
        property_name = self.property_combo.currentText()

        if plot_type == "Histogram":
            all_data = np.concatenate(all_values)
            bins = np.histogram_bin_edges(all_data, bins="auto")
            y_label = "Frequency"
            x_label = property_name
        elif plot_type == "Density":
            from scipy.stats import gaussian_kde

            all_data = np.concatenate(all_values)
            x_min, x_max = np.min(all_data), np.max(all_data)
            x_range = np.linspace(x_min, x_max, 500)
            y_label = "Density"
            x_label = property_name
        elif plot_type == "Line":
            y_label = "Value"
            x_label = "Index"
        else:
            print("Supported plot types are Histogram, Density and Line.")
            return None

        is_combined = plot_mode == "Combined" or len(data_series) == 1
        cols = 1 if is_combined else min(2, len(data_series))
        if is_combined:
            plot = self.plot_widget.addPlot(row=0, col=0)
            plot.setLabel("left", y_label)
            plot.setLabel("bottom", x_label)

            plot.disableAutoRange()
            plot.addLegend(offset=(-10, 10))

            for i, (name, obj, values, color) in enumerate(data_series):
                if plot_type == "Histogram":
                    hist, edges = np.histogram(values, bins=bins)
                    x = (edges[:-1] + edges[1:]) / 2
                    width = (edges[1] - edges[0]) * 0.8

                    if len(data_series) > 1:
                        width = width / len(data_series)
                        offset = (i - (len(data_series) - 1) / 2) * width
                    else:
                        offset = 0

                    item = pg.BarGraphItem(
                        x=x + offset,
                        height=hist,
                        width=width,
                        brush=color,
                        pen=pg.mkPen("k", width=1),
                        name=name,
                    )
                elif plot_type == "Density":
                    try:
                        kde = gaussian_kde(values)
                        density = kde(x_range)
                        item = pg.PlotDataItem(
                            x_range,
                            density,
                            pen=pg.mkPen(color, width=2),
                            fillLevel=0,
                            fillBrush=color,
                            name=name,
                        )
                    except Exception as e:
                        print(f"Error computing KDE for {name}: {e}")
                        continue
                else:
                    x = np.arange(len(values))
                    item = pg.PlotDataItem(
                        x,
                        values,
                        pen=pg.mkPen(color, width=2),
                        name=name,
                        symbol="o",
                        symbolSize=5,
                        symbolBrush=color,
                    )

                plot.addItem(item)

            plot.enableAutoRange()
            plot.autoRange()
            return None

        # For separate plots mode
        for i, (name, obj, values, color) in enumerate(data_series):
            plot = self.plot_widget.addPlot(row=i // cols, col=i % cols)
            plot.setTitle(name)
            plot.setLabel("left", y_label)
            plot.setLabel("bottom", x_label)

            if plot_type == "Histogram":
                hist, edges = np.histogram(values, bins=bins)
                x = (edges[:-1] + edges[1:]) / 2
                width = (edges[1] - edges[0]) * 0.8

                item = pg.BarGraphItem(
                    x=x,
                    height=hist,
                    width=width,
                    brush=color,
                    pen=pg.mkPen("k", width=1),
                )
            elif plot_type == "Density":
                try:
                    kde = gaussian_kde(values)
                    density = kde(x_range)
                    item = pg.PlotDataItem(
                        x_range,
                        density,
                        pen=pg.mkPen(color, width=2),
                        fillLevel=0,
                        fillBrush=color,
                    )
                except Exception as e:
                    print(f"Error computing KDE for {name}: {e}")
                    continue
            else:  # Line plot
                x = np.arange(len(values))
                item = pg.PlotDataItem(
                    x,
                    values,
                    pen=pg.mkPen(color, width=2),
                    symbol="o",
                    symbolSize=5,
                    symbolBrush=color,
                )

            plot.addItem(item)

    def _export_data(self):
        """Export analysis data to a CSV file"""
        property_name = self.property_combo.currentText()
        selected_items = self._get_selection()
        if not selected_items:
            QMessageBox.warning(
                self, "No Selection", "Please select at least one object."
            )
            return

        file_path, _ = QFileDialog.getSaveFileName(
            self, "Export Data", "", "CSV Files (*.csv);;All Files (*.*)"
        )

        if not file_path:
            return

        try:
            with open(file_path, mode="w", encoding="utf-8") as ofile:
                ofile.write(f"source,{property_name}\n")

                for name, geom in selected_items:
                    values = self.properties.get(id(geom))
                    if values is None:
                        return None

                    values = np.asarray(values).reshape(-1)
                    lines = "\n".join([f"{name},{v}" for v in values])
                    ofile.write(lines + "\n")

            QMessageBox.information(self, "Success", "Data exported successfully")

        except Exception as e:
            QMessageBox.critical(self, "Error", f"Failed to export data: {str(e)}")

    def _export_plot(self):
        """Save the current plot as an image"""
        from pyqtgraph.exporters import ImageExporter

        file_path, _ = QFileDialog.getSaveFileName(
            self, "Save Plot", "", "PNG Files (*.png);;All Files (*.*)"
        )

        if not file_path:
            return None

        try:
            exporter = ImageExporter(self.plot_widget.scene())
            exporter.parameters()["width"] = 1920
            exporter.parameters()["height"] = 1080
            exporter.parameters()["antialias"] = True
            exporter.export(file_path)
            QMessageBox.information(self, "Success", "Plot saved successfully")

        except Exception as e:
            QMessageBox.critical(self, "Error", f"Failed to save plot: {str(e)}")

    def _setup_styling(self):
        base_style = """
            QTabWidget::pane {
                border: 1px solid #cbd5e1;
                border-radius: 6px;
                top: -1px;
            }
            QTabBar::tab {
                background: transparent;
                border: 1px solid #cbd5e1;
                border-bottom: none;
                border-top-left-radius: 6px;
                border-top-right-radius: 6px;
                padding: 6px 12px;
                margin-right: 2px;
            }
            QTabBar::tab:selected {
                color: rgba(99, 102, 241, 1.0);
                border-color: rgba(99, 102, 241, 1.0);

            }
            QTabBar::tab:hover:!selected {
                color: #696c6f;
            }
            QTableWidget {
                border: 1px solid #cbd5e1;
                background-color: transparent;
                border-radius: 4px;
                outline: none;
            }
            QTableWidget QHeaderView::section {
                background-color: transparent;
                border: 1px solid #cbd5e1;
                padding: 4px;
            }
        """
        return self.setStyleSheet(base_style + QPushButton_style + QScrollArea_style)



from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import (
    QVBoxLayout,
    QDialog,
    QLabel,
    QDoubleSpinBox,
    QHBoxLayout,
    QPushButton,
    QListWidget,
    QRadioButton,
    QButtonGroup,
    QGroupBox,
    QWidget,
    QFrame,
    QMessageBox,
)

from ..widgets import DialogFooter, ContainerListWidget, StyledListWidgetItem
from ..stylesheets import (
    QGroupBox_style,
    QPushButton_style,
    QScrollArea_style,
    HelpLabel_style,
    QListWidget_style,
)


class DistanceCropDialog(QDialog):
    cropApplied = Signal(dict)

    def __init__(self, clusters, fits=[], parent=None):
        super().__init__(parent)
        self.clusters = clusters
        self.fits = fits

        self.setWindowTitle("Distance Crop")
        self.resize(800, 550)
        self.setup_ui()
        self.setStyleSheet(
            QGroupBox_style
            + QPushButton_style
            + QScrollArea_style
            + QListWidget_style
            + """
            QRadioButton {
                spacing: 5px;
            }
            QRadioButton::indicator {
                width: 18px;
                height: 18px;
            }
            """
        )

    def setup_ui(self):
        from ..icons import (
            dialog_selectall_icon,
            dialog_selectnone_icon,
        )

        main_layout = QVBoxLayout(self)
        main_layout.setSpacing(15)
        main_layout.setContentsMargins(20, 20, 20, 20)

        selection_widget = QWidget()
        selection_layout = QHBoxLayout(selection_widget)
        selection_layout.setContentsMargins(0, 0, 0, 0)
        selection_layout.setSpacing(15)

        source_panel = QGroupBox("Source Clusters")
        source_layout = QVBoxLayout(source_panel)
        source_layout.setSpacing(10)

        source_description = QLabel("Select clusters to crop based on distance")
        source_description.setStyleSheet(HelpLabel_style)
        source_layout.addWidget(source_description)

        # Quick select buttons
        quick_select_layout = QHBoxLayout()
        select_all_btn = QPushButton("Select All")
        select_all_btn.setIcon(dialog_selectall_icon)
        select_all_btn.clicked.connect(lambda: self.source_list.selectAll())
        select_none_btn = QPushButton("Clear")
        select_none_btn.setIcon(dialog_selectnone_icon)
        select_none_btn.clicked.connect(lambda: self.source_list.clearSelection())

        quick_select_layout.addWidget(select_all_btn)
        quick_select_layout.addWidget(select_none_btn)
        source_layout.addLayout(quick_select_layout)

        self.source_list = ContainerListWidget(border=False)
        self.source_list.setSelectionMode(QListWidget.SelectionMode.ExtendedSelection)
        source_layout.addWidget(self.source_list)

        selection_layout.addWidget(source_panel)

        target_panel = QGroupBox("Target Reference")
        target_layout = QVBoxLayout(target_panel)
        target_layout.setSpacing(10)

        target_description = QLabel("Select reference objects to compute distances to")
        target_description.setStyleSheet(HelpLabel_style)
        target_layout.addWidget(target_description)

        target_select_layout = QHBoxLayout()
        target_all_btn = QPushButton("Select All")
        target_all_btn.setIcon(dialog_selectall_icon)
        target_all_btn.clicked.connect(lambda: self.target_list.selectAll())
        target_none_btn = QPushButton("Clear")
        target_none_btn.setIcon(dialog_selectnone_icon)
        target_none_btn.clicked.connect(lambda: self.target_list.clearSelection())
        target_select_layout.addWidget(target_all_btn)
        target_select_layout.addWidget(target_none_btn)
        target_layout.addLayout(target_select_layout)

        self.target_list = ContainerListWidget(border=False)
        self.target_list.setSelectionMode(QListWidget.SelectionMode.ExtendedSelection)
        target_layout.addWidget(self.target_list)

        selection_layout.addWidget(target_panel)
        main_layout.addWidget(selection_widget)

        settings_group = QGroupBox("Distance Settings")
        settings_layout = QHBoxLayout(settings_group)

        distance_layout = QHBoxLayout()
        distance_label = QLabel("Maximum Distance:")
        distance_label.setMinimumWidth(150)

        self.distance_input = QDoubleSpinBox()
        self.distance_input.setValue(40.0)
        self.distance_input.setRange(0.01, 9999.99)
        self.distance_input.setMaximum(float("inf"))
        self.distance_input.setDecimals(2)
        self.distance_input.setSingleStep(1.0)
        self.distance_input.setMinimumWidth(100)
        distance_layout.addWidget(distance_label)
        distance_layout.addWidget(self.distance_input)
        distance_layout.addStretch()
        settings_layout.addLayout(distance_layout)

        direction_layout = QHBoxLayout()
        direction_label = QLabel("Keep Points:")
        direction_label.setMinimumWidth(150)
        direction_layout.addWidget(direction_label)

        self.comparison_group = QButtonGroup()

        radio_container = QFrame()
        radio_layout = QHBoxLayout(radio_container)
        radio_layout.setContentsMargins(0, 0, 0, 0)
        radio_layout.setSpacing(15)

        self.smaller_radio = QRadioButton("Within Distance")
        self.smaller_radio.setChecked(True)
        self.larger_radio = QRadioButton("Outside Distance")
        self.comparison_group.addButton(self.smaller_radio)
        self.comparison_group.addButton(self.larger_radio)

        radio_layout.addWidget(self.smaller_radio)
        radio_layout.addWidget(self.larger_radio)
        radio_layout.addStretch()

        direction_layout.addWidget(radio_container)
        settings_layout.addLayout(direction_layout)
        main_layout.addWidget(settings_group)

        footer = DialogFooter(
            info_text="Points will be filtered based on their distance to the target references.",
            dialog=self,
            margin=(0, 15, 0, 0),
        )
        main_layout.addWidget(footer)

        self.populate_lists()

    def populate_lists(self):
        for name, data in self.clusters:
            item = StyledListWidgetItem(name, data.visible, data._meta.get("info"))
            item.setData(Qt.ItemDataRole.UserRole, data)
            self.source_list.addItem(item)

        for name, data in self.clusters:
            item = StyledListWidgetItem(name, data.visible, data._meta.get("info"))
            item.setData(Qt.ItemDataRole.UserRole, data)
            self.target_list.addItem(item)

        for name, data in self.fits:
            item = StyledListWidgetItem(name, data.visible, data._meta.get("info"))
            item.setData(Qt.ItemDataRole.UserRole, data)
            self.target_list.addItem(item)

    def get_results(self):
        if self.exec() == QDialog.DialogCode.Accepted:
            return self.sources, self.targets, self.distance, self.keep_smaller
        return None, None, None, None

    def accept(self):
        self.sources = [
            x.data(Qt.ItemDataRole.UserRole) for x in self.source_list.selectedItems()
        ]
        self.targets = [
            x.data(Qt.ItemDataRole.UserRole) for x in self.target_list.selectedItems()
        ]

        if not self.sources:
            return QMessageBox.warning(
                self,
                "Selection Required",
                "Please select at least one source cluster to crop.",
            )

        if not self.targets:
            return QMessageBox.warning(
                self,
                "Selection Required",
                "Please select at least one target reference.",
            )

        self.distance = self.distance_input.value()
        self.keep_smaller = self.smaller_radio.isChecked()

        crop_data = {
            "sources": self.sources,
            "targets": self.targets,
            "distance": self.distance,
            "keep_smaller": self.keep_smaller,
        }
        self.cropApplied.emit(crop_data)
        return super().accept()



"""
ColorPreviewWidget widget for visualization of color maps.

Copyright (c) 2025 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from qtpy.QtCore import Signal
from qtpy.QtGui import QColor, QPainter
from qtpy.QtWidgets import QWidget, QPushButton, QColorDialog


class ColorPreviewWidget(QWidget):
    """Widget to display color map preview"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setMinimumHeight(20)
        self.setMaximumHeight(20)
        self.colors = self.generate_gradient("viridis")

        self.colormaps = [
            "viridis",
            "plasma",
            "magma",
            "inferno",
            "cividis",
            "turbo",
            "jet",
            "coolwarm",
            "RdBu",
            "RdYlBu",
        ]

    def generate_gradient(self, cmap_name: str, n_colors: int = None):
        from ..utils import get_cmap

        cmap = get_cmap(cmap_name)

        count = cmap.N
        if n_colors is not None:
            count = min(n_colors + 1, count)

        ret = []
        for i in range(count):
            pos = int(cmap.N * i / (count - 1))
            ret.append(QColor(*(int(x * 255) for x in cmap(pos))))
        return ret

    def set_colormap(self, cmap_name, reverse=False):
        if reverse:
            cmap_name = f"{cmap_name}_r"
        self.colors = self.generate_gradient(cmap_name)
        self.update()

    def paintEvent(self, event):
        if len(self.colors) <= 0:
            return None

        painter = QPainter(self)
        width = self.width()
        height = self.height()

        color_count = len(self.colors)
        stripe_width = width / len(self.colors)
        for i, color in enumerate(self.colors):
            x_pos = int(i * stripe_width)
            next_x = int((i + 1) * stripe_width) if i < color_count - 1 else width
            rect_width = next_x - x_pos
            painter.fillRect(x_pos, 0, rect_width, height, color)


class ColorButton(QPushButton):
    """Widget to select color"""

    colorChanged = Signal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.update_color((0, 0, 0))
        self.clicked.connect(self.choose_color)

    def update_color(self, color):
        self.current_color = [int(float(c) * 255) for c in color]
        rgb = ",".join([str(x) for x in self.current_color])
        self.setStyleSheet(f"background-color: rgb({rgb})")
        self.colorChanged.emit()

    def choose_color(self):
        color = QColor(*self.current_color)
        color_dialog = QColorDialog.getColor(initial=color, parent=self)
        if color_dialog.isValid():
            color = (
                color_dialog.red() / 255,
                color_dialog.green() / 255,
                color_dialog.blue() / 255,
            )
            self.update_color(color)
        return color

    def get_color(self, uint8: bool = False):
        if uint8:
            return self.current_color
        return [x / 255 for x in self.current_color]



"""
Dialog functions used throughout the GUI.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from qtpy.QtWidgets import (
    QDialog,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QGridLayout,
    QComboBox,
    QFrame,
    QScrollArea,
    QWidget,
    QGroupBox,
)

from .operation import make_param
from ..widgets import ParameterWidget, DialogFooter
from ..stylesheets import QGroupBox_style, QPushButton_style, QScrollArea_style


class MeshEquilibrationDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle("Mesh Equilibration")
        self.resize(600, 500)

        self._operations = [
            make_param("average_edge_length", 40.0, 0, "Average edge length of mesh."),
            make_param("lower_bound", 35.0, 0, "Minimum edge length of mesh (lc1)."),
            make_param("upper_bound", 45.0, 0, "Maximumg edge length of mesh (lc0)."),
            make_param("steps", 5000, 0, "Number of minimization steps."),
            make_param("kappa_b", 300.0, 0, "Bending energy coefficient (kappa_b)."),
            make_param("kappa_a", 1e6, 0, "Area conservation coefficient (kappa_a)."),
            make_param("kappa_v", 1e6, 0, "Volume conservation coefficient (kappa_v)."),
            make_param("kappa_c", 0.0, 0, "Curvature energy coefficient (kappa_c)."),
            make_param("kappa_t", 1e5, 0, "Edge tension coefficient (kappa_t)."),
            make_param("kappa_r", 1e3, 0, "Surface repulsion coefficient (kappa_r)."),
            make_param("volume_fraction", 1.1, 0, "Fraction VN/V0."),
            make_param("area_fraction", 1.1, 0, "Fraction AN/A0."),
            make_param(
                "scaling_lower", 1.0, 0, "Lower bound for rescaled mesh edge length."
            ),
        ]

        self.parameter_widgets = {}
        self.current_values = {op[0]: op[1] for op in self._operations}

        self.setup_ui()
        self.setStyleSheet(QGroupBox_style + QPushButton_style + QScrollArea_style)

    def setup_ui(self):
        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(20, 20, 20, 20)
        main_layout.setSpacing(15)

        header_layout = QHBoxLayout()

        title_label = QLabel("Mesh Equilibration Settings")
        title_label.setStyleSheet("font-size: 14px; font-weight: 600;")
        header_layout.addWidget(title_label)

        header_layout.addStretch()

        mode_label = QLabel("Settings Mode:")
        header_layout.addWidget(mode_label)

        self.mode_selector = QComboBox()
        self.mode_selector.addItems(["Default", "Advanced"])
        self.mode_selector.currentTextChanged.connect(self.toggle_mode)
        header_layout.addWidget(self.mode_selector)

        main_layout.addLayout(header_layout)

        # Scroll area for content
        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)
        scroll_area.setFrameShape(QFrame.Shape.NoFrame)

        content_widget = QWidget()
        scroll_area.setWidget(content_widget)
        content_layout = QVBoxLayout(content_widget)
        content_layout.setContentsMargins(0, 0, 0, 0)
        content_layout.setSpacing(15)

        # Basic settings
        basic_settings = QGroupBox("Basic Settings")
        basic_layout = QGridLayout(basic_settings)
        basic_layout.setColumnStretch(0, 1)
        basic_layout.setColumnStretch(1, 1)

        # Create edge length widget
        self.edge_length_widget = self.create_parameter_widget(self._operations[0])
        self.edge_length_widget.valueChanged.connect(self.update_bounds)
        basic_layout.addWidget(self.edge_length_widget, 0, 0)

        # Add steps widget
        self.steps_widget = self.create_parameter_widget(self._operations[3])
        basic_layout.addWidget(self.steps_widget, 0, 1)

        # Add parameters visible in both modes
        self.lower_bound_widget = self.create_parameter_widget(self._operations[1])
        self.upper_bound_widget = self.create_parameter_widget(self._operations[2])
        basic_layout.addWidget(self.lower_bound_widget, 1, 0)
        basic_layout.addWidget(self.upper_bound_widget, 1, 1)

        content_layout.addWidget(basic_settings)

        # Advanced settings
        self.advanced_group = QGroupBox("Energy Coefficients")
        advanced_layout = QGridLayout(self.advanced_group)
        advanced_layout.setColumnStretch(0, 1)
        advanced_layout.setColumnStretch(1, 1)

        # Energy coefficients (kappa)
        row, col = 0, 0
        for i in range(4, 10):
            widget = self.create_parameter_widget(self._operations[i])
            advanced_layout.addWidget(widget, row, col)
            col = (col + 1) % 2
            if col == 0:
                row += 1

        content_layout.addWidget(self.advanced_group)

        # Constraints
        self.constraints_group = QGroupBox("Additional Constraints")
        constraints_layout = QGridLayout(self.constraints_group)

        for i in range(10, len(self._operations)):
            widget = self.create_parameter_widget(self._operations[i])
            constraints_layout.addWidget(widget, (i - 10) // 2, (i - 10) % 2)

        content_layout.addWidget(self.constraints_group)
        content_layout.addStretch()

        main_layout.addWidget(scroll_area)

        footer = DialogFooter(dialog=self, margin=(0, 15, 0, 0))
        main_layout.addWidget(footer)
        self.toggle_mode("Default")

    def create_parameter_widget(self, param_data):
        """Create a parameter widget from parameter data."""
        param_name, param_value, param_min, param_info = param_data

        widget = ParameterWidget(param_name, param_value, param_min, param_info, self)
        widget.valueChanged.connect(self._on_value_changed)

        self.parameter_widgets[param_name] = widget
        return widget

    def toggle_mode(self, mode):
        """Toggle between Default and Advanced modes."""
        is_advanced = mode == "Advanced"
        self.advanced_group.setVisible(is_advanced)
        self.constraints_group.setVisible(is_advanced)

    def update_bounds(self, param_name, value):
        """Update lower and upper bounds when edge length changes."""
        if param_name != "average_edge_length":
            return

        try:
            val = float(value)
            self.lower_bound_widget.setValue(val * 0.75)
            self.upper_bound_widget.setValue(val * 1.25)

            self.current_values["lower_bound"] = val * 0.75
            self.current_values["upper_bound"] = val * 1.25
        except ValueError:
            pass

    def _on_value_changed(self, param_name, value):
        self.current_values[param_name] = value

    def get_parameters(self):
        """Get the current parameters."""
        return self.current_values



import importlib

_module_map = {
    ".animate": ["ExportManager"],
    ".interactor": ["DataContainerInteractor"],
    ".data": ["MosaicData"],
    ".container": ["DataContainer"],
    ".__version__": ["__version__"],
}

_lazy_imports = {}
for module_path, functions in _module_map.items():
    for func_name in functions:
        _lazy_imports[func_name] = (module_path, func_name)


def __getattr__(name):
    module_path, attr_name = _lazy_imports.get(name, ("", ""))

    if not module_path:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")

    mod = importlib.import_module(module_path, __name__)
    if attr_name:
        mod = getattr(mod, attr_name)

    globals()[name] = mod
    return mod



from qtpy.QtCore import Signal
from qtpy.QtWidgets import (
    QVBoxLayout,
    QLabel,
    QDoubleSpinBox,
    QSpinBox,
    QWidget,
)

from ..stylesheets import HelpLabel_style


class ParameterWidget(QWidget):
    """A widget that displays a parameter with a label and input field."""

    valueChanged = Signal(str, object)

    def __init__(
        self, param_name, param_value, param_min, param_description, parent=None
    ):
        super().__init__(parent)
        self.param_name = param_name

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(4)

        # Label with description tooltip
        label = QLabel(param_description.get("title", param_name))
        label.setToolTip(param_description.get("description", ""))
        layout.addWidget(label)

        # Help text
        if "description" in param_description:
            help_label = QLabel(param_description["description"])
            help_label.setStyleSheet(HelpLabel_style)
            help_label.setWordWrap(True)
            layout.addWidget(help_label)

        # Input field based on type
        if isinstance(param_value, float):
            self.input = QDoubleSpinBox()
            self.input.setDecimals(2)
            self.input.setRange(param_min, 1e10)
            self.input.setValue(param_value)
            self.input.setSingleStep(1.0 if param_value > 10 else 0.1)
            self.input.valueChanged.connect(
                lambda value: self.valueChanged.emit(param_name, value)
            )
        elif isinstance(param_value, int):
            self.input = QSpinBox()
            self.input.setRange(param_min, 1000000)
            self.input.setValue(param_value)
            self.input.valueChanged.connect(
                lambda value: self.valueChanged.emit(param_name, value)
            )
        else:
            self.input = QLabel(str(param_value))

        layout.addWidget(self.input)

    def getValue(self):
        if isinstance(self.input, (QDoubleSpinBox, QSpinBox)):
            return self.input.value()
        return None

    def setValue(self, value):
        if isinstance(self.input, (QDoubleSpinBox, QSpinBox)):
            self.input.setValue(value)

    def setVisible(self, visible):
        super().setVisible(visible)
        self.input.setVisible(visible)



from .legend import LegendWidget
from .scale import ScaleBarWidget
from .path_selector import PathSelector
from .dialog_footer import DialogFooter
from .sidebar import ObjectBrowserSidebar
from .mapped_combobox import MappedComboBox
from .histogram_widget import HistogramWidget
from .parameter_widget import ParameterWidget
from .trajectory_player import TrajectoryPlayer
from .ribbon import RibbonToolBar, create_button
from .color_preview import ColorPreviewWidget, ColorButton
from .volume_viewer import MultiVolumeViewer, VolumeViewer
from .coordinate_system import AxesWidget, BoundingBoxManager
from .container_list import ContainerListWidget, StyledListWidgetItem
from .status_indicator import StatusIndicator, CursorModeHandler, ViewerModes
from .settings import *



"""
Implements ProgressDialog, wrapping a ProgressBar in a dialog window.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QVBoxLayout,
    QFrame,
    QProgressBar,
    QLabel,
    QDialog,
    QApplication,
    QMessageBox,
)


class ProgressDialog:
    def __init__(self, iterable, title="Processing", parent=None):
        self.total = len(iterable)
        self.iterator = iter(iterable)

        self.current = 0
        self.dialog = QDialog(parent)
        self.dialog.setWindowTitle(title)
        self.dialog.setWindowFlags(
            Qt.WindowType.Dialog | Qt.WindowType.FramelessWindowHint
        )
        self.dialog.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)

        self.setup_ui()
        self.dialog.show()

    def setup_ui(self):
        from ..stylesheets import QProgressBar_style

        layout = QVBoxLayout(self.dialog)

        container = QFrame()
        container.setStyleSheet(
            """
            QFrame {
                background-color: white;
                border-radius: 10px;
                border: 1px solid #e5e7eb;
            }
        """
        )
        container_layout = QVBoxLayout(container)
        container_layout.setSpacing(15)

        self.progress_bar = QProgressBar()
        self.progress_bar.setTextVisible(False)
        self.progress_bar.setStyleSheet(QProgressBar_style)
        self.progress_bar.setMaximum(self.total)

        self.status_label = QLabel()
        self.status_label.setStyleSheet(
            """
            QLabel {
                color: #374151;
                font-size: 13px;
            }
        """
        )
        self.status_label.setAlignment(Qt.AlignmentFlag.AlignCenter)

        container_layout.addWidget(self.status_label)
        container_layout.addWidget(self.progress_bar)

        layout.setContentsMargins(10, 10, 10, 10)
        layout.addWidget(container)
        self.dialog.setFixedSize(300, 100)

    def update_progress(self):
        self.progress_bar.setValue(self.current)
        self.status_label.setText(f"Processing {self.current}/{self.total}")
        QApplication.processEvents()

    def __iter__(self):
        return self

    def __next__(self):
        try:
            self.update_progress()
            self.current += 1
            item = next(self.iterator)
            return item
        except StopIteration:
            self.update_progress()
            self.dialog.close()
            raise

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.dialog.close()
        if exc_type is not None:
            QMessageBox.warning(None, "Error", str(exc_value))
            # Omit traceback
            return True



from functools import partial

import numpy as np
from qtpy.QtWidgets import QWidget, QVBoxLayout, QFileDialog

from .. import meshing
from ..parallel import submit_task
from ..widgets.ribbon import create_button


def on_fit_complete(self, *args, **kwargs):
    self.cdata.data.render()
    self.cdata.models.render()


def _fit(method, geometry, **kwargs):
    from ..parametrization import PARAMETRIZATION_TYPE

    fit_object = PARAMETRIZATION_TYPE.get(method)
    if fit_object is None:
        raise ValueError(f"{method} is not supported ({PARAMETRIZATION_TYPE.keys()}).")

    points, *_ = geometry.get_point_data()

    n = points.shape[0]
    if n < 50 and method not in ["convexhull", "spline"]:
        raise ValueError(f"Insufficient points for fit ({n}<50).")
    return fit_object.fit(points, **kwargs)


def _remesh(method, geometry, **kwargs):
    import pyfqmr
    from ..meshing.utils import to_open3d
    from ..parametrization import TriangularMesh

    mesh = geometry.model
    mesh = meshing.to_open3d(mesh.vertices.copy(), mesh.triangles.copy())
    if method == "edge length":
        mesh = meshing.remesh(mesh=mesh, **kwargs)
    elif method == "vertex clustering":
        mesh = mesh.simplify_vertex_clustering(**kwargs)
    elif method == "subdivide":
        func = mesh.subdivide_midpoint
        if kwargs.get("smooth"):
            func = mesh.subdivide_loop
        kwargs = {k: v for k, v in kwargs.items() if k != "smooth"}
        mesh = func(**kwargs)
    else:
        method = kwargs.get("decimation_method", "Triangle Count").lower()
        sampling = kwargs.get("sampling")
        if method == "reduction factor":
            sampling = np.asarray(mesh.triangles).shape[0] // sampling

        simplifier = pyfqmr.Simplify()
        simplifier.setMesh(np.asarray(mesh.vertices), np.asarray(mesh.triangles))
        simplifier.simplify_mesh(
            target_count=int(sampling),
            aggressiveness=5.5,
            preserve_border=True,
            verbose=False,
        )

        vertices, faces, normals = simplifier.getMesh()
        mesh = to_open3d(vertices, faces)
    return TriangularMesh(mesh)


def _project(
    mesh_geometry,
    geometries,
    use_normals: bool = False,
    invert_normals: bool = False,
    update_normals: bool = False,
):
    from ..geometry import Geometry

    mesh = mesh_geometry.model
    new_geometries, projections, triangles = [], [], []
    for geometry in geometries:
        normals = geometry.normals if use_normals else None
        if normals is not None:
            normals = normals * (-1 if invert_normals else 1)

        kwargs = {
            "points": geometry.points,
            "normals": normals,
            "return_projection": True,
            "return_indices": False,
            "return_triangles": True,
        }
        _, projection, triangle = mesh.compute_distance(**kwargs)

        normals = geometry.normals
        if update_normals:
            normals = mesh.compute_normal(projection)

        projections.append(projection)
        triangles.append(triangle)
        new_geometries.append(
            Geometry(
                points=projection, normals=normals, sampling_rate=geometry.sampling_rate
            )
        )

    if not len(projections):
        return None

    projections = np.concatenate(projections)
    triangles = np.concatenate(triangles)
    new_mesh = mesh.add_projections(projections, triangles, return_indices=False)

    return new_mesh, new_geometries


def _run_marching_cubes(filename, **kwargs):
    from ..formats.parser import load_density
    from ..parametrization import TriangularMesh

    mesh_paths = meshing.mesh_volume(filename, **kwargs)
    sampling = load_density(filename, use_memmap=True).sampling_rate

    meshes = [TriangularMesh.from_file(x) for x in mesh_paths]
    return meshes, sampling


class ModelTab(QWidget):
    def __init__(self, cdata, ribbon, legend, **kwargs):
        super().__init__()
        self.cdata = cdata
        self.ribbon = ribbon
        self.legend = legend

        layout = QVBoxLayout(self)
        layout.setSpacing(5)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.ribbon)

    def show_ribbon(self):
        self.ribbon.clear()

        func = self._fit_parallel
        fitting_actions = [
            create_button("Sphere", "mdi.circle", self, partial(func, "sphere")),
            create_button("Ellipse", "mdi.ellipse", self, partial(func, "ellipsoid")),
            create_button("Cylinder", "mdi.hexagon", self, partial(func, "cylinder")),
            create_button(
                "RBF", "mdi.grid", self, partial(func, "rbf"), "Fit RBF", RBF_SETTINGS
            ),
            create_button(
                "Mesh", "mdi.triangle-outline", self, func, "Fit Mesh", MESH_SETTINGS
            ),
            create_button(
                "Curve",
                "mdi.chart-bell-curve",
                self,
                partial(func, "spline"),
                "Fit Spline",
                SPLINE_SETTINGS,
            ),
        ]
        self.ribbon.add_section("Parametrization", fitting_actions)

        mesh_actions = [
            create_button(
                "Sample",
                "mdi.chart-scatter-plot",
                self,
                self._sample_parallel,
                "Sample from Fit",
                SAMPLE_SETTINGS,
            ),
            create_button("To Cluster", "mdi.plus", self, self._to_cluster),
            create_button("Remove", "fa5s.trash", self, self.cdata.models.remove),
        ]
        self.ribbon.add_section("Sampling", mesh_actions)

        mesh_actions = [
            create_button(
                "Merge", "mdi.merge", self, self._merge_meshes, "Merge Meshes"
            ),
            create_button(
                "Volume",
                "mdi.cube-outline",
                self,
                self._mesh_volume,
                "Mesh Volume",
                MESHVOLUME_SETTINGS,
            ),
            create_button(
                "Repair",
                "mdi.auto-fix",
                self,
                self._repair_mesh,
                "Repair Mesh",
                REPAIR_SETTINGS,
            ),
            create_button(
                "Remesh",
                "mdi.repeat",
                self,
                self._remesh_parallel,
                "Remesh Mesh",
                REMESH_SETTINGS,
            ),
            create_button(
                "Project",
                "mdi.vector-curve",
                self,
                self._project_on_mesh,
                "Project on Mesh",
                PROJECTION_SETTINGS,
            ),
            # create_button("Skeleton", "mdi.vector-line", self, self._sceleton),
        ]
        self.ribbon.add_section("Mesh Operations", mesh_actions)

    def _to_cluster(self, *args, **kwargs):
        for geometry in self.cdata.models.get_selected_geometries():
            fit = geometry.model
            normals, sampling = None, geometry._sampling_rate
            if hasattr(fit, "mesh"):
                points = fit.vertices
                normals = fit.compute_vertex_normals()
            else:
                points = geometry.points
                if fit is not None:
                    normals = fit.compute_normal(points)

            normals = fit.compute_normal(points)

            self.cdata.data.add(points, normals=normals, sampling_rate=sampling)
        self.cdata.data.data_changed.emit()
        self.cdata.data.render()
        return None

    def _get_selected_meshes(self):
        from ..parametrization import TriangularMesh

        ret = []
        for geometry in self.cdata.models.get_selected_geometries():
            fit = geometry.model
            if not isinstance(fit, TriangularMesh):
                continue
            ret.append(geometry)
        return ret

    def _repair_mesh(
        self,
        max_hole_size=-1,
        elastic_weight=0,
        curvature_weight=0,
        volume_weight=0,
        boundary_ring=0,
        **kwargs,
    ):
        from ..parametrization import TriangularMesh

        for geometry in self._get_selected_meshes():
            fit = geometry.model
            if not hasattr(fit, "vertices"):
                continue

            fit.mesh.remove_non_manifold_edges()
            fit.mesh.remove_degenerate_triangles()
            fit.mesh.remove_duplicated_triangles()
            fit.mesh.remove_unreferenced_vertices()
            fit.mesh.remove_duplicated_vertices()
            vs, fs = meshing.triangulate_refine_fair(
                vs=fit.vertices,
                fs=fit.triangles,
                alpha=elastic_weight,
                beta=curvature_weight,
                gamma=volume_weight,
                hole_len_thr=max_hole_size,
                n_ring=boundary_ring,
            )
            self.cdata._add_fit(
                fit=TriangularMesh(meshing.to_open3d(vs, fs)),
                sampling_rate=geometry.sampling_rate,
            )

        return self.cdata.models.render()

    def _merge_meshes(self):
        from ..parametrization import TriangularMesh

        meshes, selected_meshes = [], self._get_selected_meshes()

        if len(selected_meshes) < 2:
            return None

        sampling_rate = 1
        for geometry in selected_meshes:
            sampling_rate = np.maximum(sampling_rate, geometry.sampling_rate)
            meshes.append(geometry.model)

        vertices, faces = meshing.merge_meshes(
            vertices=[x.vertices for x in meshes],
            faces=[x.triangles for x in meshes],
        )
        self.cdata._add_fit(
            fit=TriangularMesh(meshing.to_open3d(vertices, faces)),
            sampling_rate=sampling_rate,
        )
        self.cdata._models.remove(selected_meshes)

        self.cdata.models.data_changed.emit()
        return self.cdata.models.render()

    def _sceleton(self):
        selected_meshes = self._get_selected_meshes()

        if len(selected_meshes) == 0:
            return None

        for geometry in selected_meshes:
            import trimesh
            import skeletor as sk
            from ..utils import com_cluster_points

            mesh = geometry.model
            mesh = trimesh.Trimesh(mesh.vertices, mesh.triangles)
            mesh = sk.pre.fix_mesh(mesh)
            skel = sk.skeletonize.by_wavefront(mesh, waves=5, step_size=1)

            vertices = com_cluster_points(skel.vertices, 100)
            vertices = skel.vertices
            self.cdata._data.add(vertices)

        self.cdata.data.data_changed.emit()
        return self.cdata.data.render()

    def _fit_parallel(self, method: str, *args, **kwargs):
        _conversion = {
            "Alpha Shape": "convexhull",
            "Ball Pivoting": "mesh",
            "Poisson": "poissonmesh",
            "Cluster Ball Pivoting": "clusterballpivoting",
            "Flying Edges": "flyingedges",
        }
        method = _conversion.get(method, method)

        if method == "mesh":
            radii = kwargs.get("radii", None)
            try:
                kwargs["radii"] = [float(x) for x in radii.split(",")]
            except Exception as e:
                raise ValueError(f"Incorrect radius specification {radii}.") from e

        for geometry in self.cdata.data.get_selected_geometries():
            kwargs["voxel_size"] = np.max(geometry.sampling_rate)

            if method == "flyingedges" and kwargs.get("distance", -1) != -1:
                kwargs["voxel_size"] = kwargs.get("distance")

            def _callback(fit):
                self.cdata._add_fit(fit, sampling_rate=kwargs["voxel_size"])
                self.cdata.models.render()

            submit_task(
                "Parametrization",
                _fit,
                _callback,
                method.lower(),
                geometry,
                **kwargs,
            )

    def _sample_parallel(self, sampling, sampling_method, normal_offset=0.0, **kwargs):
        from ..operations import GeometryOperations

        def _callback(*args, **kwargs):
            self.cdata.data.add(*args, **kwargs)
            self.cdata.data.render()

        for geometry in self.cdata.models.get_selected_geometries():
            submit_task(
                "Sample Fit",
                GeometryOperations.sample,
                _callback,
                geometry,
                sampling,
                sampling_method,
                normal_offset,
                **kwargs,
            )

    def _remesh_parallel(self, method, **kwargs):
        selected_meshes = self._get_selected_meshes()
        if len(selected_meshes) == 0:
            return None

        method = method.lower()
        supported = (
            "edge length",
            "vertex clustering",
            "quadratic decimation",
            "subdivide",
        )
        if method not in (supported):
            raise ValueError(f"{method} is not supported, chose one of {supported}.")

        for geometry in selected_meshes:

            def _callback(fit):
                self.cdata._add_fit(fit, sampling_rate=geometry.sampling_rate)
                self.cdata.models.render()

            submit_task("Remesh", _remesh, _callback, method, geometry, **kwargs)

    def _project_on_mesh(
        self,
        use_normals: bool = False,
        invert_normals: bool = False,
        update_normals: bool = False,
        **kwargs,
    ):
        selected_meshes = self._get_selected_meshes()
        if len(selected_meshes) != 1:
            raise ValueError("Please select one mesh for projection.")

        mesh = selected_meshes[0]
        if mesh.model is None:
            return None

        def _callback(ret):
            new_mesh, new_geometries = ret

            for new_geometry in new_geometries:
                self.cdata.data.add(new_geometry)

            self.cdata._add_fit(new_mesh, sampling_rate=mesh.sampling_rate)
            self.cdata.data.render()
            self.cdata.models.render()

        submit_task(
            "Project",
            _project,
            _callback,
            mesh,
            self.cdata.data.get_selected_geometries(),
            use_normals,
            invert_normals,
            update_normals,
        )

    def _mesh_volume(self, **kwargs):
        filename, _ = QFileDialog.getOpenFileName(self, "Select Meshes")
        if not filename:
            return -1

        def _callback(ret):
            meshes, sampling = ret
            for mesh in meshes:
                self.cdata._add_fit(
                    fit=mesh,
                    sampling_rate=sampling,
                )
            self.cdata.models.render()

        submit_task(
            "Mesh Volume",
            _run_marching_cubes,
            _callback,
            filename,
        )


SAMPLE_SETTINGS = {
    "title": "Sample Fit",
    "settings": [
        {
            "label": "Sampling Method",
            "parameter": "sampling_method",
            "type": "select",
            "options": ["Points", "Distance"],
            "default": "Points",
            "notes": "Number of points or average distance between points.",
        },
        {
            "label": "Sampling",
            "parameter": "sampling",
            "type": "float",
            "min": 1,
            "default": 1000,
            "notes": "Numerical value for sampling method.",
        },
        {
            "label": "Offset",
            "parameter": "normal_offset",
            "type": "float",
            "default": 0,
            "min": -1e32,
            "notes": "Points are shifted by n times normal vector for particle picking.",
        },
    ],
}

RBF_SETTINGS = {
    "title": "RBF Settings",
    "settings": [
        {
            "label": "Direction",
            "parameter": "direction",
            "type": "select",
            "options": ["xy", "xz", "yz"],
            "default": "xy",
            "description": "Coordinate plane to fit RBF in.",
        },
    ],
}

SPLINE_SETTINGS = {
    "title": "Curve Settings",
    "settings": [
        {
            "label": "Order",
            "parameter": "order",
            "type": "number",
            "default": 3,
            "min": 1,
            "max": 5,
            "description": "Spline order to fit to control points.",
        },
    ],
}

REPAIR_SETTINGS = {
    "title": "Repair Settings",
    "settings": [
        {
            "label": "Elastic Weight",
            "parameter": "elastic_weight",
            "type": "float",
            "default": 0.0,
            "min": -(2**28),
            "description": "Control mesh smoothness and elasticity.",
            "notes": "0 - strong anchoring, 1 - no anchoring, > 1 repulsion.",
        },
        {
            "label": "Curvature Weight",
            "parameter": "curvature_weight",
            "type": "float",
            "default": 0.0,
            "min": -(2**28),
            "description": "Controls propagation of mesh curvature.",
        },
        {
            "label": "Volume Weight",
            "parameter": "volume_weight",
            "type": "float",
            "default": 0.0,
            "min": -(2**28),
            "description": "Controls internal pressure of mesh.",
        },
        {
            "label": "Boundary Ring",
            "parameter": "boundary_ring",
            "type": "number",
            "default": 0,
            "description": "Also optimize n-ring vertices for ill-defined boundaries.",
        },
        {
            "label": "Flexibility",
            "parameter": "anchoring",
            "type": "float_list",
            "default": "1,0",
            "min": "0,0",
            "max": "1,0",
            "description": "Flexibility of inferred vertices. 1 is maximum. Can be "
            "specified for all axes, e.g., 1, or per-axis, e.g., 1;1;0.5.",
        },
        {
            "label": "Hole Size",
            "parameter": "max_hole_size",
            "type": "float",
            "min": -1.0,
            "default": -1.0,
            "description": "Maximum surface area of holes considered for triangulation.",
        },
    ],
}


REMESH_SETTINGS = {
    "title": "Remesh Settings",
    "settings": [
        {
            "label": "Method",
            "parameter": "method",
            "type": "select",
            "options": [
                "Edge Length",
                "Vertex Clustering",
                "Quadratic Decimation",
                "Subdivide",
            ],
            "default": "Edge Length",
        },
    ],
    "method_settings": {
        "Edge Length": [
            {
                "label": "Edge Length",
                "parameter": "target_edge_length",
                "type": "float",
                "default": 40.0,
                "min": 1e-6,
                "description": "Average edge length to remesh to.",
            },
            {
                "label": "Iterations",
                "parameter": "n_iter",
                "type": "number",
                "default": 100,
                "min": 1,
                "description": "Number of remeshing operations to repeat on the mesh.",
            },
            {
                "label": "Mesh Angle",
                "parameter": "featuredeg",
                "type": "float",
                "default": 30.0,
                "min": 0.0,
                "description": "Minimum angle between faces to preserve the edge feature.",
            },
        ],
        "Vertex Clustering": [
            {
                "label": "Radius",
                "parameter": "voxel_size",
                "type": "float",
                "default": 40.0,
                "min": 1e-6,
                "description": "Radius within which vertices are clustered.",
            },
        ],
        "Quadratic Decimation": [
            {
                "label": "Method",
                "parameter": "decimation_method",
                "type": "select",
                "options": ["Triangle Count", "Reduction Factor"],
                "default": "Triangle Count",
                "description": "Choose how to specify the decimation target.",
            },
            {
                "label": "Sampling",
                "parameter": "sampling",
                "type": "float",
                "default": 1000,
                "min": 0,
                "description": "Numerical value for reduction method.",
            },
        ],
        "Subdivide": [
            {
                "label": "Iterations",
                "parameter": "number_of_iterations",
                "type": "number",
                "default": 1,
                "min": 1,
                "description": "Number of iterations.",
                "notes": "A single iteration splits each triangle into four triangles.",
            },
            {
                "label": "Smooth",
                "parameter": "smooth",
                "type": "boolean",
                "default": True,
                "description": "Perform smooth midpoint division.",
            },
        ],
    },
}

MESH_SETTINGS = {
    "title": "Mesh Settings",
    "settings": [
        {
            "label": "Method",
            "parameter": "method",
            "type": "select",
            "options": [
                "Alpha Shape",
                "Ball Pivoting",
                "Cluster Ball Pivoting",
                "Poisson",
                "Flying Edges",
            ],
            "default": "Alpha Shape",
        },
        *REPAIR_SETTINGS["settings"][:5],
    ],
    "method_settings": {
        "Alpha Shape": [
            {
                "label": "Alpha",
                "parameter": "alpha",
                "type": "float",
                "default": 1.0,
                "description": "Alpha-shape parameter.",
                "notes": "Large values yield coarser features.",
            },
            {
                "label": "Scaling Factor",
                "parameter": "resampling_factor",
                "type": "float",
                "default": 12.0,
                "description": "Resample mesh to scaling factor times sampling rate.",
                "notes": "Decrease for creating smoother pressurized meshes.",
            },
            {
                "label": "Distance",
                "parameter": "distance_cutoff",
                "type": "float",
                "default": 2.0,
                "description": "Vertices further than distance time sampling rate are "
                "labled as inferred for subsequent optimization.",
            },
        ],
        "Ball Pivoting": [
            {
                "label": "Radii",
                "parameter": "radii",
                "type": "text",
                "default": "50",
                "description": "Ball radii used for surface reconstruction.",
                "notes": "Use commas to specify multiple radii, e.g. '50,30.5,10.0'.",
            },
            REPAIR_SETTINGS["settings"][-1],
            {
                "label": "Downsample",
                "parameter": "downsample_input",
                "type": "boolean",
                "default": True,
                "description": "Thin input point cloud to core.",
            },
            {
                "label": "Smoothing Steps",
                "parameter": "n_smoothing",
                "type": "number",
                "default": 5,
                "description": "Pre-smoothing steps before fairing.",
                "notes": "Improves repair but less impactful for topolgoy than weights.",
            },
            {
                "label": "Neighbors",
                "parameter": "k_neighbors",
                "type": "number",
                "min": 1,
                "default": 15,
                "description": "Number of neighbors for normal estimations.",
                "notes": "Consider decreasing this value for small point clouds.",
            },
        ],
        "Cluster Ball Pivoting": [
            {
                "label": "Radius",
                "parameter": "radius",
                "type": "float",
                "default": 0.0,
                "max": 100,
                "min": 0.0,
                "description": "Ball radius compared to point cloud box size.",
                "notes": "Default 0 corresponds to an automatically determined radius.",
            },
            {
                "label": "Mesh Angle",
                "parameter": "creasethr",
                "type": "float",
                "min": 0,
                "default": 90.0,
                "description": "Maximum crease angle before stoping ball pivoting.",
            },
            {
                "label": "Smooth Iter",
                "parameter": "smooth_iter",
                "type": "number",
                "min": 1,
                "default": 1,
                "description": "Number of smoothing iterations for normal estimation.",
            },
            {
                "label": "Distance",
                "parameter": "deldist",
                "type": "float",
                "min": -1.0,
                "default": -1.0,
                "description": "Drop vertices distant from input sample points.",
                "notes": "This is post-normalization by the sampling rate.",
            },
            {
                "label": "Neighbors",
                "parameter": "k_neighbors",
                "type": "number",
                "min": 1,
                "default": 15,
                "description": "Number of neighbors for normal estimations.",
                "notes": "Consider decreasing this value for small point clouds.",
            },
        ],
        "Poisson": [
            {
                "label": "Depth",
                "parameter": "depth",
                "type": "number",
                "min": 1,
                "default": 9,
                "description": "Depth of the Octree for surface reconstruction.",
            },
            {
                "label": "Samples",
                "parameter": "samplespernode",
                "type": "float",
                "min": 0,
                "default": 5.0,
                "description": "Minimum number of points per octree node.",
            },
            {
                "label": "Smooth Iter",
                "parameter": "smooth_iter",
                "type": "number",
                "min": 1,
                "default": 1,
                "description": "Number of smoothing iterations for normal estimation.",
            },
            {
                "label": "Pointweight",
                "parameter": "pointweight",
                "type": "float",
                "min": 0,
                "default": 0.1,
                "description": "Interpolation weight of point samples.",
            },
            {
                "label": "Scale",
                "parameter": "scale",
                "type": "float",
                "min": 0,
                "default": 1.2,
                "description": "Ratio between reconstruction and sample cube.",
            },
            {
                "label": "Distance",
                "parameter": "deldist",
                "type": "float",
                "min": -1.0,
                "default": -1.0,
                "description": "Drop vertices distant from input sample points.",
                "notes": "This is post-normalization by the sampling rate.",
            },
            {
                "label": "Neighbors",
                "parameter": "k_neighbors",
                "type": "number",
                "min": 1,
                "default": 15,
                "description": "Number of neighbors for normal estimations.",
                "notes": "Consider decreasing this value for small point clouds.",
            },
        ],
        "Flying Edges": [
            {
                "label": "Distance",
                "parameter": "distance",
                "type": "float",
                "description": "Distance between points to be considered connected.",
                "default": -1.0,
                "min": -1.0,
                "max": 1e32,
                "notes": "Defaults to the sampling rate of the object.",
            },
        ],
    },
}


MESHVOLUME_SETTINGS = {
    "title": "Meshing Settings",
    "settings": [
        {
            "label": "Simplifcation Factor",
            "parameter": "simplification_factor",
            "type": "number",
            "default": 100,
            "min": 1,
            "description": "Reduce initial mesh by x times the number of triangles.",
        },
        {
            "label": "Workers",
            "parameter": "num_workers",
            "type": "number",
            "default": 8,
            "min": 1,
            "description": "Number of parallel workers to use.",
        },
        {
            "label": "Close Dataset Edges",
            "parameter": "closed_dataset_edges",
            "type": "boolean",
            "default": True,
            "description": "Close mesh at at dataset edges.",
        },
    ],
}


PROJECTION_SETTINGS = {
    "title": "Projection Settings",
    "settings": [
        {
            "label": "Cast Normals",
            "parameter": "use_normals",
            "type": "boolean",
            "default": True,
            "description": "Include normal vectors in raycasting.",
        },
        {
            "label": "Invert Normals",
            "parameter": "invert_normals",
            "type": "boolean",
            "default": False,
            "description": "Invert direction of normal vectors.",
        },
        {
            "label": "Update Normals",
            "parameter": "update_normals",
            "type": "boolean",
            "default": False,
            "description": "Update normal vectors of projection based on the mesh.",
        },
    ],
}



from typing import List

from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QDialog,
    QVBoxLayout,
    QHBoxLayout,
    QPushButton,
    QWidget,
    QListWidget,
    QGroupBox,
)

from ..widgets import DialogFooter
from ..stylesheets import QPushButton_style, QScrollArea_style


class ObjectSelectionWidget(QWidget):
    """Reusable widget for selecting objects (clusters and models)"""

    def __init__(self, cdata, parent=None):
        super().__init__(parent)
        self.cdata = cdata
        self._setup_ui()
        self.populate_lists()

        self.setStyleSheet(QPushButton_style + QScrollArea_style)

    def _setup_ui(self):
        from ..widgets import ContainerListWidget
        from ..icons import dialog_selectall_icon, dialog_selectnone_icon

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        # Quick select buttons
        objects_panel = QGroupBox("Objects")

        objects_layout = QVBoxLayout()

        quick_select_layout = QHBoxLayout()
        select_all_btn = QPushButton("Select All")
        select_all_btn.setIcon(dialog_selectall_icon)
        select_all_btn.clicked.connect(lambda: self.objects_list.selectAll())

        clear_btn = QPushButton("Clear")
        clear_btn.setIcon(dialog_selectnone_icon)
        clear_btn.clicked.connect(lambda: self.objects_list.clearSelection())

        quick_select_layout.addWidget(select_all_btn)
        quick_select_layout.addWidget(clear_btn)
        objects_layout.addLayout(quick_select_layout)

        self.objects_list = ContainerListWidget(border=False)
        self.objects_list.setSelectionMode(QListWidget.SelectionMode.ExtendedSelection)
        objects_layout.addWidget(self.objects_list)
        objects_panel.setLayout(objects_layout)

        layout.addWidget(objects_panel)

    def _make_items(self, data, data_type: str, ids: List[int]):
        from ..widgets import StyledListWidgetItem

        all_ids, new_items = [], []
        for name, obj in data:
            object_id = id(obj)

            all_ids.append(object_id)

            item = self.get_item_by_id(object_id)

            is_new = item is None
            if is_new:
                item = StyledListWidgetItem(name, obj.visible, obj._meta.get("info"))

            # Sync properties of new and existing items
            item.setText(name)
            item.set_visible(obj.visible)
            item.setData(Qt.ItemDataRole.UserRole, obj)
            item.setData(Qt.ItemDataRole.UserRole + 1, data_type)
            item.setData(Qt.ItemDataRole.UserRole + 2, id(obj))

            if is_new:
                new_items.append(item)
        return all_ids, new_items

    def populate_lists(self):
        ids = [item.data(Qt.ItemDataRole.UserRole + 2) for item in self.allItems()]

        selected = [
            item.data(Qt.ItemDataRole.UserRole + 2) for item in self.selectedItems()
        ]

        data = self.cdata.format_datalist("data")
        all_ids, new_items = self._make_items(data, "cluster", ids)

        data = self.cdata.format_datalist("models")
        _all_ids, _new_items = self._make_items(data, "models", ids)

        all_ids.extend(_all_ids)
        new_items.extend(_new_items)

        # Remove objects that have been deleted, add new items and keep selection
        deleted_ids = set(ids) - set(all_ids)
        for i in reversed(range(self.objects_list.count())):
            item = self.objects_list.item(i)
            if item.data(Qt.ItemDataRole.UserRole + 2) in deleted_ids:
                _ = self.objects_list.takeItem(i)

        for item in new_items:
            self.objects_list.addItem(item)
        return self.set_selection(set(selected) & set(all_ids))

    def get_item_by_id(self, obj_id: int):
        for i in range(self.objects_list.count()):
            item = self.objects_list.item(i)
            if item.data(Qt.ItemDataRole.UserRole + 2) == obj_id:
                return item
        return None

    def selectedItems(self):
        return self.objects_list.selectedItems()

    def allItems(self):
        return [self.objects_list.item(i) for i in range(self.objects_list.count())]

    def get_selected_objects(self):
        """Get names of selected objects"""
        return [item.text() for item in self.objects_list.selectedItems()]

    def set_selection(self, object_ids):
        """Set which objects are selected by their IDs"""
        for i in range(self.objects_list.count()):
            item = self.objects_list.item(i)
            object_id = item.data(Qt.ItemDataRole.UserRole + 2)
            item.setSelected(object_id in object_ids)


class ActorSelectionDialog(QDialog):
    """Simple dialog for selecting actors for visibility animation"""

    def __init__(self, cdata, current_selection=None, parent=None):
        super().__init__()
        self.cdata = cdata
        self.current_selection = current_selection or []

        self.setWindowTitle("Select Objects for Animation")
        self.resize(400, 450)
        self.setModal(True)

        self._setup_ui()
        self.setStyleSheet(QPushButton_style + QScrollArea_style)

    def _setup_ui(self):
        layout = QVBoxLayout(self)

        self.selection_widget = ObjectSelectionWidget(self.cdata, self)
        layout.addWidget(self.selection_widget)

        if self.current_selection:
            self.selection_widget.set_selection(self.current_selection)

        footer = DialogFooter(
            dialog=self,
            margin=(0, 15, 0, 0),
        )
        layout.addWidget(footer)

    def get_selected_objects(self):
        """Get names of selected objects"""
        return [
            item.data(Qt.ItemDataRole.UserRole + 2)
            for item in self.selection_widget.selectedItems()
        ]



from typing import Dict

from ._utils import get_extension, CompatibilityUnpickler
from .parser import (
    read_star,
    read_tsv,
    read_txt,
    read_mesh,
    read_volume,
    read_tsi,
    read_vtu,
    read_structure,
    GeometryDataContainer,
)

FORMAT_MAPPING = {
    read_star: ["star"],
    read_tsv: ["tsv"],
    read_txt: ["txt", "xyz", "csv"],
    read_mesh: ["obj", "ply", "stl", "off", "gltf", "glb," "fbx"],
    read_volume: ["mrc", "em", "map", "h5", "mrc.gz", "em.gz", "map.gz"],
    read_tsi: ["q", "tsi"],
    read_vtu: ["vtu"],
    read_structure: ["pdb", "cif", "gro"],
}


def open_file(filename: str, *args, **kwargs) -> GeometryDataContainer:
    """
    Open and parse a file based on its extension.

    Parameters
    ----------
    filename : str
        Path to the file to be opened.
    *args
        Additional positional arguments passed to the parser function.
    **kwargs
        Additional keyword arguments passed to the parser function.

    Returns
    -------
    GeometryDataContainer
        Parsed geometry data container.

    Raises
    ------
    ValueError
        If the file extension is not supported.
    """
    extension = get_extension(filename)[1:]

    func = None
    for reader_func, reader_formats in FORMAT_MAPPING.items():
        if extension not in reader_formats:
            continue
        func = reader_func

    if func is None:
        supported = ", ".join([f"'{x}'" for t in FORMAT_MAPPING.values() for x in t])
        raise ValueError(f"Unknown extension '{extension}', supported are {supported}.")
    return func(filename, *args, **kwargs)


def open_session(filename: str, *args, **kwargs) -> Dict:
    """
    Open and deserialize a pickled session file.

    Parameters
    ----------
    filename : str
        Path to the session file to be opened.
    *args
        Additional positional arguments (unused).
    **kwargs
        Additional keyword arguments (unused).

    Returns
    -------
    Dict
        Deserialized session data.
    """
    with open(filename, "rb") as ifile:
        unpickler = CompatibilityUnpickler(ifile)
        data = unpickler.load()
    return data



import qtawesome as qta

icon_color = "#4f46e5"
dialog_margin = (16, 16, 16, 16)
footer_margin = (0, 16, 0, 0)

dialog_accept_icon = qta.icon("mdi.chevron-right", color=icon_color)
dialog_reject_icon = qta.icon("mdi.close", color=icon_color)

dialog_next_icon = qta.icon("mdi6.skip-next", color=icon_color)
dialog_previous_icon = qta.icon("mdi6.skip-previous", color=icon_color)
dialog_apply_icon = qta.icon("mdi6.check-all", color=icon_color)

dialog_selectall_icon = qta.icon("mdi.select-all", color=icon_color)
dialog_selectnone_icon = qta.icon("mdi6.select-remove", color=icon_color)

info_icon = qta.icon("mdi.information-outline", color=icon_color).pixmap(18, 18)

cluster_icon = qta.icon("mdi.chart-bubble", color=icon_color)
model_icon = qta.icon("mdi.shape", color=icon_color)
cluster_icon = None
model_icon = None
